/*****************************************************************************/
/*                                                                           */
/*                              SOFTING AG                                   */
/*                        Richard-Reitzner-Allee 6                           */
/*                              D-85540 Haar                                 */
/*                      Phone: (++49)-(0)89-45656-0                          */
/*                      Fax:   (++49)-(0)89-45656-399                        */
/*                                                                           */
/*                  Copyright (C) SOFTING AG 1995-2007                       */
/*                                                                           */
/* See file CREDITS for list of people who contributed to this               */
/* project.                                                                  */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,                    */
/* MA 02111-1307, USA.                                                       */
/*                                                                           */
/* email: support.automation@softing.com                                     */
/* Softing AG, Richard-Reitzner-Allee 6, 85540 Haar, Germany                 */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/
/*                                                                           */
/*  Author      : SOFTING                                                    */
/*                                                                           */
/*  CHANGE_NOTES                                                             */
/*                                                                           */
/*  date      name      change                                               */
/*  -----------------------------------------------------------------------  */
/*****************************************************************************/
/*****************************************************************************/

/***************************************************************************/
/*                                                                         */
/*          LINUX PROFIBUS DEVICE DRIVER FOR SOFTING PROFIBOARD            */
/*                                                                         */
/* The data structure CtrlExt has entries for 4 possible boards. For each  */
/* board the hw-resources are stored. In addition this data structure      */
/* contains sub structures for (direct) pointers to the DP-Ram memory      */
/* of the different services (for example start address of DP input        */
/* and output areas).                                                      */
/*                                                                         */
/* For DP two bottom half interrupt service threads exist. Only one        */
/* of them is used, they are switched - if necessary - in the ISR.         */
/*                                                                         */
/* The device driver uses _NOT_ a dynamic major number. A fixed value has  */
/* been choosen (122). It has been written, in mind, that several boards   */
/* have to be supported (later) and that it can be ported to other         */
/* hardware platforms.                                                     */
/*                                                                         */
/* Data device: The driver offers two access modes to the data device:     */
/*              - the normal mode and the                                  */
/*              - optimized mode.                                          */
/*              In the NORMAL mode a read call to the data device returns  */
/*              immideatly with the requested data.                        */
/*              In the OPTIMIZED mode a read call to the data device       */
/*              will only return immideatly, if the data image had been    */
/*              changed since the last successful read call. In case the   */
/*              device had been opened in blocking mode, the calling       */
/*              thread/process will be blocked, until real new data is     */
/*              available, in case of non-blocking mode the function       */
/*              returns immideatly but without any data (count=0).         */
/*              The OPTIMIZED mode can only be enabled globally. This      */
/*              means, that if the data device is opened several times     */
/*              and via one file descriptor the device is set to the       */
/*              OPTIMIZED mode, this mode is used for every fd!            */
/*                                                                         */
/* init_module:                                                            */
/*   - searching of HW (board)                                             */
/*   - allocation of HW-resources                                          */
/*   - initializing of CtrlExt                                             */
/*                                                                         */
/* open:                                                                   */
/*   - increment InUseCount (for the specific board and access mode)       */
/*   - for the first time:                                                 */
/*     + init DP interrupt service thread (bottom half)                    */
/*                                                                         */
/* close                                                                   */
/*   - decrement Board->InUseCount (specific for board and access mode)    */
/*   - free memory                                                         */
/*   - in case of board device / write mode -> get new FWVersion (proc)    */
/*                                                                         */
/*                                                                         */
/* ----                                                                    */
/* A good resource for information about writing Linux-Device-Drivers is:  */
/* Alessandro Rubini: LINUX Device Drivers, O'Reilly & Associates, INC,    */
/*                    February 1998                                        */
/***************************************************************************/

#if defined(PB_ISA_SUPPORT) && defined (PB_PCI_SUPPORT)
#error this driver does not support mixing of isa and pci boards
#endif

#include <linux/module.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,33)
#include <linux/autoconf.h>
#endif
#include <asm/io.h>
#include <asm/segment.h>


#include "kernel_version.h"

#include <linux/poll.h>


#include <linux/slab.h>
#include <linux/proc_fs.h>

#include "basetypes.h"
#include "pb_debug.h"
#include "cntrl.h"
#include "cmi.h"
#include "xx_if.h"
#include "xx_err.h"
#include "pb_xxx.h"
#include "fw.h"
#include "pb_ioctl.h"

#if defined(CONFIG_MODVERSION) && !defined(MODVERSION)
#define MODVERSIONS
#endif
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#ifdef DEBUG
  #undef MOD_DEC_USE_COUNT
  #undef MOD_INC_USE_COUNT

  #define MOD_DEC_USE_COUNT
  #define MOD_INC_USE_COUNT
#endif

#if (VERSION_CODE(2,4,0) <= LINUX_VERSION_CODE)
MODULE_LICENSE ("GPL");
#endif

#define PB_STATISTICS


#define PB_HW_CONFIG_TIMEOUT  1 * HZ/100


#define ENABLE_IRQ_CHANNEL_0      0x00000001
#define ENABLE_IRQ_CHANNEL_1      0x00000008
#define DISABLE_IRQ_CHANNEL_0     (~ENABLE_IRQ_CHANNEL_0)
#define DISABLE_IRQ_CHANNEL_1     (~ENABLE_IRQ_CHANNEL_1)

#define RUN_CHANNEL_0             0x00000020
#define RUN_CHANNEL_1             0x00000800
#define RESET_CHANNEL_0           (~RUN_CHANNEL_0)
#define RESET_CHANNEL_1           (~RUN_CHANNEL_1)

#define SET_IRQ_CHANNEL_0         0x00000004
#define SET_IRQ_CHANNEL_1         0x00000100
#define CLEAR_IRQ_CHANNEL_0       (~SET_IRQ_CHANNEL_0)
#define CLEAR_IRQ_CHANNEL_1       (~SET_IRQ_CHANNEL_1)


/***************************************/
/* Definition of Minornumbers          */
/* Bit 0-3: Select the Board (0-15)    */
/* Bit 4-6: 000 = Board Device         */
/*          001 = Service Device       */
/*          010 = Data Device          */
/*          011 = Slave Input Device   */
/*          100 = Slave Output Device  */
/* Bit 7:   reserved                   */
/***************************************/
#define PB_BOARD_SELECT   0x0F
#define PB_DEVICE_SELECT  0xF0

#define PB_BOARD_DEVICE     0x0
#define PB_SERVICE_DEVICE   0x10
#define PB_DATA_DEVICE      0x20
#define PB_SLAVE_IN_DEVICE  0x30
#define PB_SLAVE_OUT_DEVICE 0x40


#define PRINT(a,b) 




// Device Table for CAN-ACx-PCI (subsystem Id 7 und 8)


// DECLARE_MUTEX (globalMutex);

typedef unsigned char uchar;

struct semaphore PbGlobalSema;

struct CntrlExt           CtrlExt[PB_MAX_NR_BOARDS];



unsigned long dummy;



extern int cmi_init( struct CntrlExt *CtrlExt );
extern int cmi_init_slave( struct CntrlExt *CtrlExt );
unsigned short reset_cntrl_irq( struct CntrlExt *CtrlExt);
extern int PbWriteBoard( struct CntrlExt *Board, char const *buffer,int count );
extern void PbStopChannel(struct CntrlExt *Board);

extern BOOL isBufferFree(void);
extern BOOL copyToFwBuffer(u8 *pBuffer, u32 size);
extern u32 doFirmwareDownload(struct CntrlExt  *pCtrlExt);



void PbGetFwVersion(char *pDpram, USIGN32 *pVersion, USIGN32 *pSubversion);
int boot_handle_sbn_record(struct CntrlExt *Board, const char *pBuffer, long size);
ssize_t pb_read( struct file *file, char *buf, size_t count, loff_t *unused);
ssize_t pb_write(struct file *file, const char *buf, size_t count, loff_t *unused);
unsigned int pb_poll (struct file *file, poll_table *table);
int pb_open(struct inode *Inode, struct file *file);
int pb_close(struct inode *Inode, struct file *file);





MODULE_DESCRIPTION("Softing Profibus driver");
MODULE_VERSION("Version 1.6.1");


extern void StatisticTimer( unsigned long arg );

struct file_operations PrFileOps = 
{
  .llseek = default_llseek,
  .read = pb_read,
  .write = pb_write,
  .poll = pb_poll,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
    .unlocked_ioctl = pb_ioctl,
#else
    .ioctl = pb_ioctl,
#endif
  .open = pb_open,
  .release = pb_close,

  .owner = THIS_MODULE,

};

// proc file system
extern struct proc_dir_entry profiboard_entry;

long pb_sleep_on_timeout (long timeout, WAIT_STRUCT *wait_reason)
{
  long sleeps = (timeout>0) ? timeout : 0;
  volatile int counter;
  
  if(timeout>60000) timeout = 60000;
  
  // printk("pb_sleep_on_timeout ...\n");
  if (timeout > 0) 
  {
    get_wait(wait_reason);
    counter = wait_reason->counter;
    while((sleeps > 0) && (counter != 0))
    {
      sleeps--;
      udelay(1000);
      counter = wait_reason->counter;
    }
        
    if(wait_reason->counter != 0)
    {
      // printk("pb_sleep_on_timeout done!\n");
      return( -EIO );
    }
  } 
  else 
  {
    get_wait(wait_reason);
    while(wait_reason->counter != 0)
    {
      mdelay(1);
    }    
    timeout = 0;
  }
  if (signal_pending (current)) 
  {
    printk("profiboard(data): ERESTARTSYS\n");
    return( -ERESTARTSYS );
  }
  
  // printk("pb_sleep_on_timeout done!\n");
  return (timeout);
}

void leaveGlobMutex(void)
{
  // printk("leaveGlobMutex\n");
  up(&PbGlobalSema);  
//   printk("leaveGlobMutex ok\n");
//   up(&globalMutex);
}

int enterGlobMutex(void)
{
  int ret;
//   return down_interruptible(&globalMutex);
//   printk("enterGlobMutex\n");
  ret = down_interruptible(&PbGlobalSema);
  if(ret)
  {
    printk("enterGlobMutex error ***************************************************** \n");
  }
  
  return ret;
}

static unsigned long spEntries[20];
static int num = 0;
// saves all lines from which the spinlock is called.
void checkEntries(unsigned long line)
{
  int i;
  
  if(num>=20) 
  {
    for(i=0; i<num; i++)
    {
      printk("%lu, ", spEntries[i]);
    }
    printk("\n");
    
    return;
  }
  else
  { 
    for(i=0; i<num; i++)
    {
      if(line==spEntries[i])
      {
	for(i=0; i<num; i++)
	{
// 	  printk("%lu, ", spEntries[i]);
	}
	printk("\n");
	
	return;
      }
    }
    // not found
    spEntries[num] = line;
    
    for(i=0; i<=num; i++)
    {
      printk("%lu, ", spEntries[i]);
    }
    printk("\n");
    
    num++;
  }
  
  return;
}

void GetSpinlock(struct CntrlExt *Board, unsigned long *Flags, unsigned long line)
{
  
  unsigned long IrqFlags = 0;
//   int belegt;
  static unsigned long oldline=0;

//   printk("GetSpinlock: counter %i, line %lu\n",Board->insideSpinlock,line);
//   belegt = (Board->insideSpinlock>0);
//   if(belegt)
//   {
//     printk("GetSpinlock: spinlock war belegt; line=%u oldline=%u\n",line,oldline);
//   }
  
  spin_lock_irqsave(&Board->Spinlock, IrqFlags);
//   printk("GetSpinlock: ok, counter %i, line %lu\n",Board->insideSpinlock,line);
  Board->insideSpinlock++;
//   if(belegt)
//   {
//     printk("GetSpinlock: spinlock war belegt; line=%lu oldline=%lu\n",line,oldline);
//   }
  
  if(Board->insideSpinlock != 1)
  {
//     printk("GetSpinlock: ** spinlock error ***%i,%lu\n",Board->insideSpinlock,line);
  }
  
//   checkEntries(line);
  
  
  oldline = line;
  
  *Flags = IrqFlags;
}

void ReleaseSpinlock(struct CntrlExt *Board, unsigned long Flags, unsigned long line)
{
   Board->insideSpinlock--;
  if(Board->insideSpinlock)
  {
    
    printk("ReleaseSpinlock: ** spinlock error ***********************%i,%lu\n",Board->insideSpinlock,line);
  }
//   printk("ReleaseSpinlock: counter %i, line %lu\n",Board->insideSpinlock,line);
  spin_unlock_irqrestore(&Board->Spinlock, Flags);
//   spin_unlock_irq(&Board->Spinlock);

  
}



void StartTimeout( int *Timeout, int value)
{
  *Timeout = value;
}

// Upps, we really sleep here
int SleepAWhile( int *Timeout )
{
  (*Timeout)--;
  udelay( 800 );
  if( *Timeout<0 )
    return( 1 );
  else
    return( 0 );
}

void reset_my_wait(WAIT_STRUCT *pWait)
{
  pWait->counter = 0;
}

void init_my_wait(WAIT_STRUCT *pWait, char *name)
{
  pWait->counter = 0;
  pWait->name = name;
  spin_lock_init(&pWait->lock);
}

void get_wait(WAIT_STRUCT *pWait)
{
  unsigned long irqflags;
  
  if((pWait->counter<-1) || (pWait->counter > 1))
  { 
    // should never happen!
    printk("get_wait : counter = %4i %s\n",pWait->counter, pWait->name);
  }
  
  // printk("get_wait: %s\n",pWait->name);
  spin_lock_irqsave(&pWait->lock, irqflags);

  pWait->counter++;
  if((pWait->counter<-1) || (pWait->counter > 1))
  { // should never happen!
    printk("get_wait : counter = %4i %s\n",pWait->counter, pWait->name);
  }
  spin_unlock_irqrestore(&pWait->lock, irqflags);
//   spin_unlock_irq(&pWait->lock);
}

void free_wait(WAIT_STRUCT *pWait)
{
  unsigned long irqflags;
  
  spin_lock_irqsave(&pWait->lock, irqflags);
//   spin_lock_irq(&pWait->lock);
  pWait->counter--;
    
  // printk("free_wait: %s\n",pWait->name);
  
  if(pWait->counter == -2)
  {
    printk("free_wait: counter = %4i %s is not used!\n",pWait->counter, pWait->name);
  }
  spin_unlock_irqrestore(&pWait->lock, irqflags);
//   spin_unlock_irq(&pWait->lock);
  // printk("free_wait: ok, %s\n",pWait->name);
}


void DumpASCII(unsigned char* pChar, unsigned long size, unsigned long offset)
{
  unsigned long i;
  
  if(offset < size)
    return;
  
  for(i=0; i<size; i++)
  {
    if((pChar[i] >= ' ') && (pChar[i] <= 127))
    {
      printk("%c",pChar[i]);
    }
    else
    {
      printk(".");
    }
    
  }
    
}

void DumpMem(void *pMem, unsigned long size, char *pText)
{
  unsigned long i, lineLength = 0x10;
  unsigned char *pVal = pMem;
  
  printk("************************* %s *******************************\n",pText);
  printk("length=%lu\n",size);
  
  for(i=0; i<size; i++)
  {
    if(!(i%lineLength))
    {
      DumpASCII(&pVal[i-lineLength], lineLength, i);
      printk("\n0x%08lx: ",i+(PB_ADDRESS)pMem);
    }
    
    printk("%02x ",pVal[i]);
    
  }
  printk("\n");
  
}



/********/
/* read */
/********/

struct inode i;
char intbuf[1000];

ssize_t pb_read( struct file *file,
  char *buf, size_t count, loff_t *unused)

{
  unsigned short Length  = (unsigned short) count;
  int DevType            = GET_MINOR_TYPE ();
  struct CntrlExt *Board = (struct CntrlExt *)file->private_data;
  long timeout           = 0;
  int i, offset; // used for data device
  short Result;
  unsigned long flags;
  
  //printk("pb_read:\n");
  
  
  
  if(enterGlobMutex())
  {
    return (-EINTR);
  }
  
  // printk("pb_read: DevType=0x%x, board %u\n",DevType,Board->Channel);
  
  if( DevType == PB_DATA_DEVICE ) 
  {
    /***************/
    /* Data Device */
    /***************/
    Board->Data.Calls++;
    offset = file->f_pos - Board->Dp.InOffset;
    //printk("offset=%u\n",offset);
    if( count > Board->Dp.InLength ) 
    {
      count = Board->Dp.InLength;
    }
    PR_DEBUG(D_DATA_L1, "pb_read %d bytes from %ld\n",
        count, (long) file->f_pos );
    if( (count+offset) > Board->Dp.InLength ) 
    {
      count = Board->Dp.InLength - offset;
      if( count < 0 )
        count = 0;
    }
    if( Board->Data.OptimizedMode ) 
    {
      // avoid the race condition, that the data arrives between
      // checking and going to sleep.
     // cli();
     // printk("optimized mode\n");
      
      GetSpinlock(Board, &flags, __LINE__);
      if( !Board->Data.NewInputData ) 
      {
        // no changes in the data
        if( !(file->f_flags & O_NONBLOCK) ) 
        {
          /*****************/
          /* BLOCKING MODE */
          /*****************/
          ReleaseSpinlock(Board, flags, __LINE__);
          // printk("%s line %u\n",__func__, __LINE__);
          if ((timeout = pb_sleep_on_timeout 
               (
                 Board->timeout,
                &Board->Data.wait_for_new_input_data
               )
              ) < 0
             ) 
          {
           PRINT("pb_read: timeout line:%u\n",__LINE__);
            // printk("pb_read: return\n");
            leaveGlobMutex();
            return (timeout);
          }
          
        } 
        else 
        {
          /*********************/
          /* NON BLOCKING MODE */
          /*********************/
          // no data available
          count=0;
          ReleaseSpinlock(Board, flags, __LINE__);
          
        }
      }
      else
      {
        ReleaseSpinlock(Board, flags, __LINE__); // kno 21.01.12
      }
    }
 
    if( Board->Dp.InBuf && count ) 
    {
      GetSpinlock(Board, &flags, __LINE__);
      // This is done because a call of copy_to_user is not allowed inside a spinlock!
      memcpy(Board->Dp.InBufCopy, Board->Dp.InBuf+offset, count);
      Board->Data.NewInputData = 0;
      ReleaseSpinlock(Board, flags, __LINE__);
      
      dummy = copy_to_user( buf, Board->Dp.InBufCopy, count );
      
      printk("InBufCopy: %p: 0x%02x 0x%02x 0x%02x 0x%02x \n", 
             &Board->Dp.InBufCopy[0],
             ((unsigned char *)(Board->Dp.InBufCopy))[0], 
             ((unsigned char *)(Board->Dp.InBufCopy))[1], 
             ((unsigned char *)(Board->Dp.InBufCopy))[2], 
             ((unsigned char *)(Board->Dp.InBufCopy))[3]);
      
    } 
    else 
    {
      count=0;
    }
    Length = count;
#ifdef DEBUG
    // just for debugging!
    if( Board->Dp.StateBuf ) 
    {
      for( i=0; i<127; i++ ) 
      {
        // mask 1. bit -> master
        if( ((char*)Board->Dp.StateBuf)[i] & 0x7f )
          PR_DEBUG(D_DATA_L1,"State Station %d -> %x\n", i,
              ((unsigned char*)Board->Dp.StateBuf)[i] );
      }
    }
#endif
  } 
  else 
    if( DevType == PB_SERVICE_DEVICE ) 
  {
    /******************/
    /* Service Device */
    /******************/
    PR_DEBUG(D_SERVICE_L1,"profiboard: read(%d) pid=%d\n",
             Length,current->pid );
    
   // printk("profiboard: read(%d) pid=%d\n",Length,current->pid );
    
    
    // To prevent a race condition we disable interrupts here!
    // In between identifying, that we have to wait for the data
    // (Length==0), and setting ourselve to "sleep", the interrupt
    // might occure and we miss it :-(   --> therefore we have to
    // disable the interrupts. (interruptible_sleep_on will release
    // interrupts again)
    Result = cmi_read( Board, buf, &Length );
    Board->Service.CMIClosed=0;
    // In blocking mode we have to wait for data
    if( Length==0 ) 
    {
      if( Board->timeout == 0 ) 
      { 
        // Windows NT non blocking mode
//        printk("profiboard(read): NT-Blocking mode return\n");
       // printk("pb_read: return\n");
       // sti();
        leaveGlobMutex();
        return( -EIO );
      } // Board->timeout
      
      if( !(file->f_flags & O_NONBLOCK) ) 
      {
        /*****************/
        /* blocking mode */
        /*****************/
        timeout = Board->timeout;
       // cli();
        GetSpinlock(Board, &flags, __LINE__);
        while( CmiCheckForTelegram(Board) != CON_IND_RECEIVED ) 
        {
          PR_DEBUG(D_SERVICE_L1,"blocking - timeout %ld\n", timeout );
          Length = (unsigned short) count;
          // release spinlock while sleeping; kno inserted 22.02.08
          ReleaseSpinlock(Board, flags, __LINE__);
          // printk("%s line %u\n",__func__, __LINE__);
          if ((timeout = pb_sleep_on_timeout (timeout, &Board->Service.wait_for_read)) < 0) 
          {
//             PRINT("pb_read: timeout line:%u\n",__LINE__);
            // printk("pb_read: return\n");
            leaveGlobMutex();
            return (timeout);
          }
          cmi_read( Board, buf, &Length );
          
          if( Length ) 
          { // got it
            break;
          }
          if( Board->Service.InUseCountWrite == 0)
          {
            // printk("pb_read: return\n");
            leaveGlobMutex();
            return (-EIO);
          }
        }
        //  printk("%s\n",intbuf);
      

        timeout = 0;
        set_current_timeout (timeout);
        // Ok, we have the data, return it to the user
      } 
      else 
      {
        /*********************/
        /* non blocking mode */
        /*********************/
       // PRINT("pb_read: -EAGAIN line:%u\n",__LINE__);
        leaveGlobMutex();
        //printk("pb_read: -EAGAIN line:%u\n",__LINE__);
        // printk("pb_read: return\n");
        return( -EAGAIN );
      }
    }
    if( Length ) 
    {
      T_PROFI_SERVICE_DESCR *pSdb = (T_PROFI_SERVICE_DESCR*) buf;
#ifdef PB_STATISTICS
      Board->Service.Calls++;
#endif
      if (pSdb->layer == DP_USR)
      {
         if ((pSdb->primitive == CON) && (pSdb->service == DP_ACT_PARAM_LOC))
         {
            if (Board->Dp.dpStatePending != 0)
            {
               if (pSdb->result == 0)
               {

                  if ((Board->Dp.dpStatePending == DP_OP_MODE_STOP   ) &&
                      (Board->Dp.dpState        == DP_OP_MODE_OFFLINE)    )
                  {
                     cmi_init_dp(Board);
                  }
                  Board->Dp.dpState = Board->Dp.dpStatePending;
               }
            }
            Board->Dp.dpStatePending = 0;
         }
      }
      //wake_up_interruptible( &Board->Service.QWrite );
      // printk("free_wait old wait_for_write\n");
//       free_wait(&Board->Service.wait_for_write);
    }
  } 
  else if( DevType == PB_SLAVE_IN_DEVICE ) 
  {
    if((Result=cmi_get_dps_input_data(Board,&Length,buf))!=E_OK) 
    {
      
      printk( KERN_ERR "cmi_get_dps_input_data: %d, Length: %d\n",
              Result, Length );
      Length = 0;
    }
  } 
  else 
    if( DevType == PB_SLAVE_OUT_DEVICE ) 
  {
    // Buffer has to have the size output_data_len + 1 or more.
    // Otherwise no byte will be returned.
    offset = file->f_pos;
    if( offset < 0 ) 
    {
      // printk("pb_read: return\n");
      return( -EINVAL );
    }
    if( count == 0 ) 
    {
      // printk("pb_read: return\n");
      return( 0 );
    }
    count--;
    if( count > Board->SlaveOut.Cache.output_data_len ) 
    {
      count = Board->SlaveOut.Cache.output_data_len;
    }
    if( offset > Board->SlaveOut.Cache.output_data_len ) 
    {      //  file->f_pos += Length;
        //  printk("%s\n",intbuf);
      

      offset = Board->SlaveOut.Cache.output_data_len;
    } 
    if( (count+offset) > Board->SlaveOut.Cache.output_data_len ) 
    {
      count = Board->SlaveOut.Cache.output_data_len - offset;
    }
    PR_DEBUG(D_SLAVE_L1,"pb_read: SLAVE OUT %d bytes from %ld\n",
             count, (long) file->f_pos );
    // avoid that the cache is written just when we read! --> cli()
    GetSpinlock(Board, &flags, __LINE__); //cli
    memcpy (Board->SlaveOut.Cache.output_data_copy+offset, Board->SlaveOut.Cache.output_data+offset,count);
    memcpy (
      &Board->SlaveOut.Cache.output_state_copy, 
      &Board->SlaveOut.Cache.output_state, 
      sizeof (Board->SlaveOut.Cache.output_state)); 
    
    Board->SlaveOut.Cache.output_state &=~(DPS_OUTPUT_STATE_NEW_DATA|DPS_OUTPUT_STATE_GLOBAL_CONTROL);
    ReleaseSpinlock(Board, flags, __LINE__);
    dummy = copy_to_user (buf, Board->SlaveOut.Cache.output_data_copy+offset,count);
    dummy = copy_to_user (buf+count, &Board->SlaveOut.Cache.output_state_copy, sizeof (Board->SlaveOut.Cache.output_state_copy)); 
    
    count++;
    Length = count;
		  
  } 
  else 
      if( DevType == PB_BOARD_DEVICE ) 
    { // read the firmware version
    /****************/
    /* Board Device */
    /****************/
    if( file->f_pos == 0 ) 
    {
      USIGN32 Version, Subversion;
      static int counter = 0;
      
      counter = (counter + 1) & 1;
      PbGetFwVersion((void *)Board->DprAddr, &Version, &Subversion); 
      
      if(FALSE)
      {
        if(counter)
        {
          cmi_get_data( Board, ID_FW_VERS_IMAGE, file->f_pos,
                        &Length, intbuf, PB_USER_SPACE );
        
          if(Length > count)
          { // passt nicht rein
           // printk("to long\n");
            Length = count;
            dummy = copy_to_user(buf, intbuf, Length);
            counter++;
          }
          else
          {
           // printk("ok\n");
            dummy = copy_to_user(buf, intbuf, Length);
          }    
        }    
      }
      else
      {
        char *asic[] = {"UNKNOWN","ASPC2","SPC3","ASPC2 and SPC3"};        
        sprintf(intbuf, "\nBOOT FIRMWARE V%c%c%c\n%s\nASIC: %s\n%s\n\n",
                ((char*)(Board->DprAddr + 0x75))[0],
                ((char*)(Board->DprAddr + 0x75))[1],
                ((char*)(Board->DprAddr + 0x75))[2],
                (char*)(Board->DprAddr + 0xc0), 
                asic[Board->asic&3],
                "(c)Copyright Softing Automation GmbH 1998-2013");
      
        Length = strlen(intbuf)+1;
      
      //  printk("count: %x\n",count);
     
        if(counter)
        {
          if(Length > count)
          { // passt nicht rein
           // printk("to long\n");
            Length = count;
            dummy = copy_to_user(buf, intbuf, Length);
            counter++;
          }
          else
          {
           // printk("ok\n");
            dummy = copy_to_user(buf, intbuf, Length);
          }
        } 
        else
        {
          Length = 0;
        }
      }
    }
    else 
    {
      Length = 0;
    }
  } 
  else 
  {
    Length = -ENODEV;
  }
  leaveGlobMutex();
 // printk("%u Bytes read\n",Length);
  return( Length );
}

/*********/
/* write */
/*********/

ssize_t pb_write(struct file *file,
  const char *buf, size_t count, loff_t *unused)

{
  int DevType            = GET_MINOR_TYPE ();
  struct CntrlExt *Board = (struct CntrlExt *)file->private_data;
  long   timeout         = 0;
  T_PROFI_SERVICE_DESCR sdb;
  int offset;
  unsigned long flags;
  int16 ret;
 
  
  if(enterGlobMutex())
  {
    return (-EINTR);
  }
  
  // printk("pb_write: DevType=0x%x, board %u\n",DevType,Board->Channel);
  
  // for read and write verify_area is not necessary (Rubini, page 129)
  if( DevType == PB_DATA_DEVICE ) 
  {
    /***************/
    /* Data Device */
    /***************/
    Board->Data.Calls++;
    if( count > Board->Dp.OutLength ) 
    {
      count = Board->Dp.OutLength;
    }
    //printk("pb_write %d bytes from %ld\n",count,(long)file->f_pos);
    PR_DEBUG(D_DATA_L1,"profiboard(write) %d to /dev/pbdata\n", count );
    offset = file->f_pos-Board->Dp.OutOffset;
    if( (count+offset) > Board->Dp.OutLength ) 
    {
      count = Board->Dp.OutLength - offset;
      if( count < 0 )
        count = 0;
    }
    // XXX - Is disabling IRQ's enough to avoid having two
    //       processes writing to the same image at one time?
    //cli(); 
    
    if( Board->Dp.OutBuf ) 
    {
      dummy = copy_from_user (Board->Dp.OutBufCopy, buf, count);
      GetSpinlock(Board, &flags, __LINE__);
       
      memcpy (Board->Dp.OutBuf+offset, Board->Dp.OutBufCopy, count);
      if (Board->Dp.dpState == DP_OP_MODE_STOP)
      {
         copy_to_io ((void*)Board->Dp.OutAddr,Board->Dp.OutBuf,Board->Dp.OutLength);
      }
      ReleaseSpinlock(Board, flags, __LINE__);
    }
    
  } 
  else 
    if( DevType == PB_SERVICE_DEVICE ) 
  {
    /******************/
    /* Service Device */
    /******************/
    PR_DEBUG(D_SERVICE_L1,"profiboard(write) %d bytes pid %d\n",
        count, current->pid);
        PR_DEBUG(D_SERVICE_L3,"jiffies= %ld\n", jiffies);
        
    // printk("pb_write: PB_SERVICE_DEVICE\n");
//     schedule();
    
    if( (ret=cmi_write(Board,(char *)buf,&count,PB_USER_SPACE)) != E_OK ) 
    {
      PR_DEBUG(D_SERVICE_L1,"can't issue service - got %d\n",ret);
      
      dummy = copy_from_user (&sdb, buf, sizeof(T_PROFI_SERVICE_DESCR)); // sdb is a local variable!
      
      while( sdb.layer==DP && ((ret==E_IF_NO_PARALLEL_SERVICES) ||
        (ret==E_IF_SERVICE_CONSTR_CONFLICT)) ) 
      {
        Board->Service.CMIClosed=1; // no race condition? - No.
        if( (file->f_flags & O_NONBLOCK) ) 
        { 
          // non blocking mode
          break;
        }
        timeout = 2000;       // 2sec.
        // printk("%s line %u\n",__func__, __LINE__);
        if ((timeout = pb_sleep_on_timeout (timeout, &Board->Service.wait_for_write)) < 0) 
        {
          if (timeout == -EIO) 
          {
             break;
          }
        }
        ret=cmi_write(Board,(char *)buf,&count,PB_USER_SPACE);
        
      }
    }
    
  switch( ret ) 
        {
    case E_IF_FATAL_ERROR:
      count=-E_IF_FATAL_ERROR;
      break;
    case E_IF_SERVICE_CONSTR_CONFLICT:
      // sdb.layer must be set, because either the retcode is E_OK or
      // the data has been copied from the user buffer to sdb.
      if( sdb.layer==DP )
        count=-E_IF_SERVICE_CONSTR_CONFLICT;
      break;
    case E_IF_NO_PARALLEL_SERVICES:
      // sdb.layer must be set, because either the retcode is E_OK or
      // the data has been copied from the user buffer to sdb.
      if( sdb.layer==DP )
        count=-E_IF_NO_PARALLEL_SERVICES;
      break;
    case E_OK:
#ifdef PB_STATISTICS
    Board->Service.Calls++;
#endif
    {
      T_PROFI_SERVICE_DESCR *pSdb  = (T_PROFI_SERVICE_DESCR *) &buf[0];
      USIGN8                *pData = (USIGN8*) (pSdb + 1);
      
      if ((pSdb->layer == DP) && ((pSdb->service == DP_INIT_MASTER) || (pSdb->service == DP_EXIT_MASTER)))
      {
        Board->Dp.dpState = DP_OP_MODE_OFFLINE;
        cmi_exit_dp(Board);
      } 
      else if ((pSdb->layer == DP) && (pSdb->service == DP_ACT_PARAM_LOC))
      {
        T_DP_ACT_PARAM_REQ *pActParam = (T_DP_ACT_PARAM_REQ*) pData;
        
        if (pActParam->area_code == DP_AREA_SET_MODE)
        {
          if      (pActParam->activate == DP_OP_MODE_OFFLINE)     Board->Dp.dpStatePending = DP_OP_MODE_OFFLINE;
          else if (pActParam->activate == DP_OP_MODE_STOP)        Board->Dp.dpStatePending = DP_OP_MODE_STOP;
          else if (pActParam->activate == DP_OP_MODE_CLEAR)       Board->Dp.dpStatePending = DP_OP_MODE_CLEAR;
          else if (pActParam->activate == DP_OP_MODE_RED_CLEAR)   Board->Dp.dpStatePending = DP_OP_MODE_CLEAR;
          else if (pActParam->activate == DP_OP_MODE_OPERATE)     Board->Dp.dpStatePending = DP_OP_MODE_OPERATE;
          else if (pActParam->activate == DP_OP_MODE_RED_OPERATE) Board->Dp.dpStatePending = DP_OP_MODE_OPERATE;
        }
      }
    }
    break;
    
    default:
      count=0;
      break;
        }
        //         restore_flags( flags );
    } 

    else 
      if( DevType == PB_BOARD_DEVICE ) 
      {
    /****************/
    /* Board Device */
    /****************/
    //count = 0;
   //printk("profiboard(fw_download) %d bytes\n", count );
// #ifdef DEBUG
    if( (Board->FwDwnld.TarHeader== NULL &&
      Board->FwDwnld.TarHeaderAct == NULL &&
      Board->FwDwnld.Firmware== NULL &&
      Board->FwDwnld.FirmwareAct==(char*)0L)&&
      ( count < MAX_DEBUG_CNTRL_LEN ) ) {
      char DebugCmd[MAX_DEBUG_CNTRL_LEN+1]; // for a closing 0

      dummy = copy_from_user (DebugCmd, buf, count);
      DebugCmd[count] = 0;
      DebugArea = GetDebugFromString( DebugCmd );
      PR_DEBUG(D_ALL_L1,"debuglevel = 0x%x\n", DebugArea );
     // printk("debuglevel = 0x%x\n", DebugArea );
      PR_DEBUG(D_ALL_L1,"debuglevel = 0x%x\n", DebugArea );
    } 
    else 
    {
// #else
//     {
// #endif
      DumpMem((char*)buf, count, "Firmware");
      switch(Board->DeviceType)
      {
        case SOFTING_PB_PRO_DEVICE:
        case SOFTING_PB_PRO_104_DEVICE:
	case SOFTING_PB_PCIE_DEVICE:
          
          if((!Board->isFwLoading) && !isBufferFree())
          {
            printk("parallel firmware download is not possible!\n");
          }
          else
          {
            //DumpMem((char*)buf, count, "Firmware");
            Board->isFwLoading = TRUE;
            copyToFwBuffer((char*)buf,count);
          }
          break;
          
        default:
        {
          USIGN32 Version, Subversion;
     
          PbGetFwVersion((void *)Board->DprAddr, &Version, &Subversion);  
          file->f_pos = count; 
          // printk("version=%lu, subversion=%lx\n",Version,Subversion);
     
          if((Version <= 250) && (Subversion<0x3110009))
          {
            ret = (int16) PbWriteBoard( Board, buf, count );
            if( ret<=0 ) 
            { // ready or error occured
              count = ret;
            }
          }
          else
          {
            // printk("count = %u\n", count);
            DumpMem((char*)buf, 0x100, "Firmware");
            boot_handle_sbn_record(Board, buf, count);
          }          
        }
          break;
        
      }
 
    }

  } 
  else 
  {
    count = -ENODEV;
  }

  PR_DEBUG(D_ALL_L1,"wrote %d bytes pid %d\n", count, current->pid );
  // printk("pb_write returns %i\n",count);
  leaveGlobMutex();
  return( count );
}

void scheduleTask(struct CntrlExt *Board)
{
 // printk("scheduleTask: %u\n", Board->active_tsklt);
  
  switch(Board->active_tsklt)
  {
    case PBDATATASK:
       tasklet_schedule(&Board->PbDataTasklet);
      break;
      
    case PBFIRSTTASK:
       tasklet_schedule(&Board->PbFirstTasklet);
      break;
      
    case PBSLAVETASK:
       tasklet_schedule(&Board->PbSlaveTasklet);
      break;
      
    default:
      
      break;
    
  }
}

/******************************************************************************/
/*                             OPEN                                           */
/*                                                                            */
/* Boarddevice:   n for reading, 1 for writing AND no other device is open!   */
/*                RDWR is not allowed here!                                   */
/* Servicedevice: n for writing, 1 for reading AND board not open for writing */
/* Datadevice:    1 for writing, n for reading AND board not open for writing */
/* SlaveInDevice: 1 for writing, n for reading AND board not open for writing */
/* SlaveOutDevice:1 for writing, n for reading AND board not open for writing */
/******************************************************************************/

int pb_open(struct inode *Inode, struct file *file)

{
  int ActBoard           = GET_MINOR_BOARD ();
  int DevType            = GET_MINOR_TYPE ();
  struct CntrlExt *Board = &CtrlExt[ActBoard];

 // PRINT("pb_open:\n");
  if(enterGlobMutex())
    return (-EINTR);
  
  PR_DEBUG(D_ALL_L1,"profiboard: open mode %x - %x for board %d (%p)\n",
           file->f_flags, O_NONBLOCK, ActBoard, Board );
  
//printk("profiboard: open mode %x - %x for board %d (%p)\n",file->f_flags, O_NONBLOCK, ActBoard, Board );
  
  if( Board->Board.InUseCountRead==0 && Board->Board.InUseCountWrite==0 &&
    Board->Service.InUseCountRead==0 &&Board->Service.InUseCountWrite==0 &&
    Board->Data.InUseCountRead==0 && Board->Data.InUseCountWrite==0 &&
    Board->SlaveIn.InUseCountRead==0 && Board->SlaveIn.InUseCountWrite==0 &&
    Board->SlaveOut.InUseCountRead==0&&Board->SlaveOut.InUseCountWrite==0) 
  {
    // First start of open.
    PR_DEBUG(D_ALL_L1,"first start of open\n");
    if ( (Board->port == 0) && (Board->ConfigBase == 0) ) 
    {
      leaveGlobMutex();
      return( -ENODEV );
    }

    PbStartBoard( Board, C165_RESET );
    if( Board->TypeOfFW == PB_DP_SLAVE_FW ) 
    {
      //Board->PbDataISTQueue.data = (void *)Board;
     // Board->PbDataISTQueue.routine = PbSlaveIST;
     // printk("pb_open: line %u tasklet_init(&Board->PbSlaveTasklet, PbSlaveIST, (TASKPARATYPE) Board)\n", __LINE__);
      Board->active_tsklt = PBSLAVETASK;
      
      
      cmi_init_slave( Board );
    } 
    else 
    {
     // Board->PbDataISTQueue.data = (void *)Board;
     // Board->PbDataISTQueue.routine = PbDataISTFirst;
     // printk("pb_open: line %u tasklet_init(&Board->PbFirstTasklet, PbDataISTFirst, (TASKPARATYPE)Board)\n", __LINE__);
      Board->active_tsklt = PBFIRSTTASK;
    }
    // Board->PbDataISTQueue.sync = 0; // must be initialized to zero
  }
  PR_DEBUG(D_BOARD_L1,"Board %d/%d, Srvc %d/%d, Data %d/%d SlIn %d/%d SlOut %d/%d\n",
      Board->Board.InUseCountRead, Board->Board.InUseCountWrite,
      Board->Service.InUseCountRead, Board->Service.InUseCountWrite,
      Board->Data.InUseCountRead, Board->Data.InUseCountWrite,
      Board->SlaveIn.InUseCountRead, Board->SlaveIn.InUseCountWrite,
      Board->SlaveOut.InUseCountRead, Board->SlaveOut.InUseCountWrite);
  // if the board is opened for writing --> no other access is allowed!
  if( Board->Board.InUseCountWrite ) 
  {
    leaveGlobMutex();
    return( -ENODEV );
  }

  if( DevType == PB_BOARD_DEVICE ) 
  {
    switch( file->f_flags & O_ACCMODE ) 
    {
    case O_RDONLY:
      Board->Board.InUseCountRead++;
      break;
    case O_WRONLY:
      if( Board->Board.InUseCountRead   ||
        Board->Service.InUseCountRead ||Board->Service.InUseCountWrite||
        Board->Data.InUseCountRead    ||Board->Data.InUseCountWrite   ||
        Board->SlaveIn.InUseCountRead ||Board->SlaveIn.InUseCountWrite||
        Board->SlaveOut.InUseCountRead
        ||Board->SlaveOut.InUseCountWrite ) 
      {
        leaveGlobMutex();
        return( -ENODEV );
      }
      Board->Board.InUseCountWrite++;
      break;
    case O_RDWR:
      leaveGlobMutex();
      return( -ENODEV );
    }
  } 
  else 
    if( DevType == PB_SERVICE_DEVICE ) 
  {
    // n for writing, 1 for reading
    switch( file->f_flags & O_ACCMODE ) 
    {
    case O_RDONLY:
      if( Board->Service.InUseCountRead == 0 ) 
      {
        Board->Service.InUseCountRead++;
      } else 
      {
        leaveGlobMutex();
        return( -ENODEV );
      }
      break;
    case O_WRONLY:
      Board->Service.InUseCountWrite++;
      break;
    case O_RDWR:
      if( Board->Service.InUseCountRead == 0 ) 
      {
        Board->Service.InUseCountRead++;
        Board->Service.InUseCountWrite++;
      } else 
      {
        leaveGlobMutex();
        return( -ENODEV );
      }
      break;
    }
  } 
  else if( DevType == PB_DATA_DEVICE ) 
  {
    if( Board->TypeOfFW != PB_MASTER_FW ) 
    {
      PR_DEBUG( D_ALL_L1, KERN_ERR "no master firmware!\n");
      leaveGlobMutex();
      return( -ENODEV );
    }
    // 1 for writing, n for reading
    switch( file->f_flags & O_ACCMODE ) 
    {
    case O_RDONLY:
      Board->Data.InUseCountRead++;
      break;
    case O_WRONLY:
      if( Board->Data.InUseCountWrite == 0 ) 
      {
        Board->Data.InUseCountWrite++;
      } 
      else 
      {
        leaveGlobMutex();
        return( -ENODEV );
      }
      break;
    case O_RDWR:
      if( Board->Data.InUseCountWrite == 0 ) 
      {
        Board->Data.InUseCountWrite++;
        Board->Data.InUseCountRead++;
      } 
      else 
      {
        leaveGlobMutex();
        return( -ENODEV );
      }
      break;
    }
  } 
  else if( DevType == PB_SLAVE_IN_DEVICE ) 
  {
    if( Board->TypeOfFW != PB_DP_SLAVE_FW ) 
    {
      PR_DEBUG( D_SLAVE_L1, KERN_ERR "no slave firmware!\n");
      leaveGlobMutex();
      return( -ENODEV );
    }
    switch( file->f_flags & O_ACCMODE ) 
    {
    case O_RDONLY:
      Board->SlaveIn.InUseCountRead++;
      break;
    case O_WRONLY:
      if( Board->SlaveIn.InUseCountWrite==0 ) 
      {
        Board->SlaveIn.InUseCountWrite++;
      } 
      else 
      {
        leaveGlobMutex();
        return( -ENODEV );
      }
      break;
    case O_RDWR:
      if( Board->SlaveIn.InUseCountWrite==0 ) 
      {
        Board->SlaveIn.InUseCountWrite++;
        Board->SlaveIn.InUseCountRead++;
      } 
      else 
      {
        leaveGlobMutex();
        return( -ENODEV );
      }
      break;
    }
  } 
  else if( DevType == PB_SLAVE_OUT_DEVICE ) 
  {
    if( Board->TypeOfFW != PB_DP_SLAVE_FW )
    {
      leaveGlobMutex();
      return( -ENODEV );
    }
    switch( file->f_flags & O_ACCMODE ) 
    {
    case O_RDONLY:
      Board->SlaveOut.InUseCountRead++;
      break;
    case O_WRONLY:
      if( Board->SlaveOut.InUseCountWrite==0 ) 
      {
        Board->SlaveOut.InUseCountWrite++;
      } 
      else 
      {
        leaveGlobMutex();
        return( -ENODEV );
      }
      break;
    case O_RDWR:
      if( Board->SlaveOut.InUseCountWrite==0 ) {
        Board->SlaveOut.InUseCountWrite++;
        Board->SlaveOut.InUseCountRead++;
      } 
      else 
      {
        leaveGlobMutex();
        return( -ENODEV );
      }
      break;
    }
  } 
  else 
  {
    leaveGlobMutex();
    return( -ENODEV );
  }

  file->private_data = Board;
  Board->timeout = -1; // no timeout
  MOD_INC_USE_COUNT;
  leaveGlobMutex();
  return( 0 );
}

/***************/
/* Select/Poll */
/***************/


unsigned int pb_poll (struct file *file,
  poll_table *table)
{
  struct CntrlExt *Board = file->private_data;
  int DevType            = GET_MINOR_TYPE ();
  unsigned int mask      = 0;

 // PRINT("pb_poll\n");
  PR_DEBUG (D_ALL_L1, "profiboard(poll - board %p\n", Board);
  if(enterGlobMutex())
    return (-EINTR);

  if (DevType == PB_DATA_DEVICE) 
  {
    if ((file->f_mode & FMODE_READ) && Board->Data.OptimizedMode) 
    {
      // printk("poll wait QNewInputData .. "); 
      poll_wait(file, &Board->Data.QNewInputData, table);
      // printk("poll wait QNewInputData ok\n");
      if (Board->Data.NewInputData == 0) 
      {
        // no new data available, let's go to sleep on this FD
        PR_DEBUG(D_DATA_L1,"No Data available\n");
        leaveGlobMutex();
        return (0);
      }
      leaveGlobMutex();
      return (POLLIN | POLLRDNORM);
    }
    PR_DEBUG(D_DATA_L1,"Data available\n");
     //always read- and writable
    leaveGlobMutex();
    return (POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM);
  } 
  else 
    if (DevType == PB_SERVICE_DEVICE) 
  {
    // printk("poll wait QRead .. "); 
    poll_wait(file, &Board->Service.QRead, table);
    // printk("poll wait QRead ok\n");
    // printk("poll wait QWrite .. "); 
    poll_wait(file, &Board->Service.QWrite, table);
    // printk("poll wait QWrite ok\n");
    if (file->f_mode & FMODE_READ) 
    {
      if (CmiCheckForTelegram (Board) == CON_IND_RECEIVED) 
      {
        PR_DEBUG(D_SERVICE_L1,"Gotcha!\n");
        mask |= POLLIN | POLLRDNORM;
      }
    }
    if (file->f_mode & FMODE_WRITE) 
    {
      if (Board->Service.CMIClosed) 
      {
        PR_DEBUG(D_SERVICE_L1,"select out --> going to sleep\n");
        mask |= POLLERR;
      } 
      else 
      {
        mask |= POLLOUT | POLLWRNORM;
      }
    }
    leaveGlobMutex();
    return (mask);
  } 
  else 
  {
    leaveGlobMutex();
    return (POLLERR); // does not support non blocking mode
  }
}


int CheckForBootblock( struct CntrlExt *Board )
{
  T_CMI_DESCRIPTOR *cmi;
  __u32 volatile *BootFlag;

 // PRINT("CheckForBootblock\n");
  
  cmi = (T_CMI_DESCRIPTOR *)Board->DprAddr;
  BootFlag = (__u32 *)(Board->DprAddr + OFFSET_BOOT_BLOCK_FLAG);
  //init_cntrl( Board );
  PbInitBoard( Board, 0 ); // no reset
//   write_io_long (&cmi->c_ready_mask, 0);
//   write_io_long (BootFlag, FLAG_NO_BOOT_BLOCK );
  
  cmi->c_ready_mask = 0;
  *BootFlag = FLAG_NO_BOOT_BLOCK;
  
  PbInitBoard( Board, C165_RESET );
  
  
  // now wait for the boot flag
  return( WaitForBootFlag( Board ) );
}

int WaitForBootFlag( struct CntrlExt *Board )
{
  int timeout;
  __u32 volatile *BootFlag;

  printk("WaitForBootFlag\n");
  BootFlag = (__u32 *)(Board->DprAddr + OFFSET_BOOT_BLOCK_FLAG);
  StartTimeout( &timeout, CMI_TIMEOUT );
  while( !SleepAWhile( &timeout ) ) 
  {
    if( *BootFlag == FLAG_BOOT_BLOCK ) 
    {
      PR_DEBUG( D_ALL_L1, "profiboard: bootblock found\n");
      printk("WaitForBootFlag finished\n");
      return( 1 );
    }
  }
  printk(KERN_INFO "profiboard: no boot block\n");
  DumpMem((void*)Board->DprAddr, 0x200, "DPRAM");
  return( 0 );
}






/*********/
/* CLOSE */
/*********/

int pb_close(struct inode *Inode, struct file *file)
{
  struct CntrlExt *Board = file->private_data;
  int DevType            = GET_MINOR_TYPE ();
  int *InUseCountRead    = NULL;
  int *InUseCountWrite   = NULL;

  unsigned long flags;


  PR_DEBUG(D_ALL_L1,  "profiboard: close %p\n", Board);
  if( DevType == PB_BOARD_DEVICE ) 
  {
    InUseCountRead  = &Board->Board.InUseCountRead;
    InUseCountWrite = &Board->Board.InUseCountWrite;
    
    switch(Board->DeviceType)
    {
      case SOFTING_PB_PRO_DEVICE:
      case SOFTING_PB_PRO_104_DEVICE:
      case SOFTING_PB_PCIE_DEVICE:
        if(Board->isFwLoading)
        {
          if(enterGlobMutex())
            return (-EINTR);

          doFirmwareDownload(Board);
          Board->isFwLoading = FALSE;
          
          leaveGlobMutex();
        }
        break;
        
      default:
        if( Board->FwDwnld.TarHeader ) 
        {
          kfree( Board->FwDwnld.TarHeader );
          Board->FwDwnld.TarHeader = Board->FwDwnld.TarHeaderAct = NULL;
          PR_DEBUG(D_ALL_L1, "freeing TarHeader mem\n");
        }
        if( Board->FwDwnld.Firmware ) 
        {
          kfree( Board->FwDwnld.Firmware );
          Board->FwDwnld.Firmware = Board->FwDwnld.FirmwareAct = NULL;
          PR_DEBUG(D_ALL_L1, "freeing Firmware mem\n");
        }        
        break;
      
    }
    

  } 
  else if( DevType == PB_SERVICE_DEVICE ) 
  {
    // printk("close service device\n");
    reset_my_wait(&Board->Service.wait_for_read);
    reset_my_wait(&Board->Service.wait_for_write);
    
    InUseCountRead  = &Board->Service.InUseCountRead;
    InUseCountWrite = &Board->Service.InUseCountWrite;
  } 
  else if ( DevType == PB_DATA_DEVICE ) 
  {
    // printk("close data device\n");
    reset_my_wait(&Board->Data.wait_for_new_input_data);
    InUseCountRead  = &Board->Data.InUseCountRead;
    InUseCountWrite = &Board->Data.InUseCountWrite;
    if( *InUseCountRead==0 )
      Board->Data.OptimizedMode = 0;
  } 
  else if( DevType == PB_SLAVE_IN_DEVICE ) 
  {
    // printk("close PB_SLAVE_IN_DEVICE device\n");
    reset_my_wait(&Board->Data.wait_for_new_input_data);
    reset_my_wait(&Board->Service.wait_for_read);
    reset_my_wait(&Board->Service.wait_for_write);
    InUseCountRead  = &Board->SlaveIn.InUseCountRead;
    InUseCountWrite = &Board->SlaveIn.InUseCountWrite;
  } 
  else if( DevType == PB_SLAVE_OUT_DEVICE ) 
  {
    // printk("close PB_SLAVE_OUT_DEVICE device\n");
    reset_my_wait(&Board->Data.wait_for_new_input_data);
    reset_my_wait(&Board->Service.wait_for_read);
    reset_my_wait(&Board->Service.wait_for_write);
    InUseCountRead  = &Board->SlaveOut.InUseCountRead;
    InUseCountWrite = &Board->SlaveOut.InUseCountWrite;
  }
  if( InUseCountRead && InUseCountWrite ) 
  {
    switch( file->f_flags & O_ACCMODE ) 
    {
    case O_RDONLY:
      (*InUseCountRead)--;
      break;
    case O_WRONLY:
      (*InUseCountWrite)--;
      break;
    case O_RDWR:
      (*InUseCountWrite)--;
      (*InUseCountRead)--;
      break;
    }
  } 
  else 
  {
    printk( KERN_ERR "profiboard - close: No InUseCount (%p,%p)\n",
        InUseCountRead, InUseCountWrite);
  }
  PR_DEBUG(D_BOARD_L1,"Board %d/%d, Srvc %d/%d, Data %d/%d SlIn %d/%d SlOut %d/%d\n",
      Board->Board.InUseCountRead, Board->Board.InUseCountWrite,
      Board->Service.InUseCountRead, Board->Service.InUseCountWrite,
      Board->Data.InUseCountRead, Board->Data.InUseCountWrite,
      Board->SlaveIn.InUseCountRead, Board->SlaveIn.InUseCountWrite,
      Board->SlaveOut.InUseCountRead, Board->SlaveOut.InUseCountWrite);
  // All devices are closed ...
  if( Board->Board.InUseCountRead==0 && Board->Board.InUseCountWrite==0 &&
    Board->Service.InUseCountRead==0 && Board->Service.InUseCountWrite==0 &&
    Board->Data.InUseCountRead==0 && Board->Data.InUseCountWrite==0 &&
    Board->SlaveIn.InUseCountRead==0 && Board->SlaveIn.InUseCountWrite==0 &&
    Board->SlaveOut.InUseCountRead==0&&Board->SlaveOut.InUseCountWrite==0) 
  {

    printk ("all devices closed\n");
    printk ("stop firmware - close hardware\n");
    
    cmi_exit_dp( Board );
   // save_flags(flags);
   // cli(); solved 21.2.08 kno
    
    GetSpinlock(Board, &flags, __LINE__);
#ifdef PB_PCI_SUPPORT
    PbStopChannel(Board);
#else
   // PbStartCntrl(Board);
    outb_p( 0, Board->port+BOARD_CTRL );
#endif
  ReleaseSpinlock(Board, flags, __LINE__);
 // restore_flags (flags);

  // XXX - Autsch, busy loop!
  udelay( 35 ); // IRQ has to be activated for at least 25 usec

  }
  MOD_DEC_USE_COUNT;
  return (0);

}



/*****************************************/
/* Interrupt Service Tasks (Bottom Half) */
/*****************************************/

void PbDataISTFirst( TASKPARATYPE Args )
{
  struct CntrlExt *Board = (struct CntrlExt *)Args;
  unsigned long flags;

 // printk("PbDataISTFirst:\n");
  PR_DEBUG(D_ALL_L1,"PbDataISTFirst( %xp )\n", Args );
  GetSpinlock(Board, &flags, __LINE__);
  // switch to the real copying function
  
  Board->active_tsklt = PBDATATASK;
  
  // removed 13.3.08 laut boe nicht nÃ¶tig! irq_to_cntrl( Board, DP_SLAVE_IO_ACK_IRQ );
  irq_to_cntrl( Board, DP_SLAVE_IO_REQ_IRQ ); // cycle completed
  Board->Data.NewInputData = ( Board->Dp.InLength ) ? 1 : 0;
  ReleaseSpinlock(Board, flags, __LINE__);
}

int checkData(char *pAddr)
{
  int i;
  char c;
  
  c=*pAddr;
  for(i=0; i<0x30; i++)
  {
    if((char)(c+i) != pAddr[i])
    {
      return i;
    }
  }
  return 0;
}

void checkOData(char *buffer, int sz, unsigned char *caller)
{
  int i;
  unsigned char first=(unsigned char)buffer[0];
  
  if(sz<48)
  {
    printk("data image too small!\n");
    return;
  }
  
  first=(unsigned char)buffer[0];
  for(i=0; i<16; i++,first++)
  {
    if((unsigned char)first != (unsigned char)buffer[i])
    {
      printk("Error in byte %u, 0x%x!=0x%x, called form %s\n", i, (unsigned char)buffer[i],first,caller);
    }
  }
  
  first=(unsigned char)buffer[16];
  for(i=16; i<32; i++,first++)
  {
    if((unsigned char)first != (unsigned char)buffer[i])
    {
      printk("Error in byte %u, 0x%x!=0x%x, called form %s\n", i, (unsigned char)buffer[i],first,caller);
    }
  }
  
  first=(unsigned char)buffer[32];
  for(i=32; i<48; i++,first++)
  {
    if((unsigned char)first != (unsigned char)buffer[i])
    {
      printk("Error in byte %u, 0x%x!=0x%x, called form %s\n", i, (unsigned char)buffer[i],first,caller);
    }
  }
  
}

void PbDataIST( TASKPARATYPE Args )
{
  unsigned long flags;
  struct CntrlExt *Board = (struct CntrlExt *)Args;

 // printk("PbDataIST:\n");
  Board->Data.Cycles++;
  GetSpinlock(Board, &flags, __LINE__);
  if (Board->Dp.OutBuf) 
  {
    copy_to_io ((void*)Board->Dp.OutAddr,Board->Dp.OutBuf,Board->Dp.OutLength);
//     checkOData((char*)Board->Dp.OutAddr, Board->Dp.OutLength, __func__);
  }
  if (Board->Dp.StateBuf) 
  {
    copy_from_io (Board->Dp.StateBuf, (void*) Board->Dp.StateAddr, Board->Dp.StateLength);
  }
  /* irq_to_cntrl( Board, DP_SLAVE_IO_ACK_IRQ );  ACK_IRQ not needed */
  if( Board->Dp.InBuf ) 
  {
    // for the optimized dp-mode, test, whether the data changed or not
    // The following check about the mode is not necessary, but the
    // driver is faster in normal mode operation of the data device.
    if (Board->Data.OptimizedMode) 
    {
      if (compare_io_kernel ((void*)Board->Dp.InAddr,Board->Dp.InBuf, Board->Dp.InLength)) 
      {
        PR_DEBUG(D_SLAVE_L1,"Data changed!\n");
        Board->Data.NewInputData = 1;
        wake_up_interruptible( &Board->Data.QNewInputData );
        free_wait(&Board->Data.wait_for_new_input_data);
      }
    }
   // DumpMem((void*) (Board->Dp.InAddr+48), 0x30, "DPRAM inputs");
    copy_from_io (Board->Dp.InBuf, (unsigned char*)Board->Dp.InAddr, Board->Dp.InLength);
  }
  // cycle completed
  irq_to_cntrl( Board, DP_SLAVE_IO_REQ_IRQ );
  ReleaseSpinlock(Board, flags, __LINE__);
//   printk("InAddr: 0x%lx: 0x%02x 0x%02x 0x%02x 0x%02x \n", 
//          Board->Dp.InAddr,
//          ((unsigned char *)(Board->Dp.InAddr))[0], 
//          ((unsigned char *)(Board->Dp.InAddr))[1], 
//          ((unsigned char *)(Board->Dp.InAddr))[2], 
//          ((unsigned char *)(Board->Dp.InAddr))[3]);
//   
//     printk("InBuf: %p: 0x%02x 0x%02x 0x%02x 0x%02x \n", 
//          &Board->Dp.InBuf[0],
//          ((unsigned char *)(Board->Dp.InBuf))[0], 
//          ((unsigned char *)(Board->Dp.InBuf))[1], 
//          ((unsigned char *)(Board->Dp.InBuf))[2], 
//          ((unsigned char *)(Board->Dp.InBuf))[3]);
}

void PbSlaveIST( TASKPARATYPE Args )
{
  struct CntrlExt *Board = (struct CntrlExt *)Args;
  unsigned long flags;

 // printk("PbSlaveIST:\n");
  Board->SlaveOut.Cycles++;
  GetSpinlock(Board, &flags, __LINE__);
  cmi_get_dps_output_data( Board );
  
  irq_to_cntrl( Board, DPS_SLAVE_OUTP_ACK_IRQ );
  ReleaseSpinlock(Board, flags, __LINE__);
}

/**************************************************************************/
/*                      INTERRUPT SERVICE ROUTINE                         */
/*                                                                        */
/* The DP stack will send a DP_SLAVE_IO_REQ_IRQ when changing the state   */
/* from STOP to CLEAR. This IRQ has to be acknowledged by sending         */
/* an IRQ back to the controller from the same type. The same IRQ occures */
/* everytime a DP cycle is finished.                                      */
/* By the open call the driver is initialized using a "primary" IST       */
/* (PbDataISTFirst), which prepares the datastructure for the data        */
/* copying routine, allocates the kernel memory buffer and switches       */
/* to the "productive" IST, which is doing in the following all           */
/* necessary data copying.                                                */
/*                                                                        */
/* The interrupt service procedure consists of two steps. The ISR         */
/* is called first (interrupts are disabled), which calls the IST,        */
/* which is called on Linux "bottom-half".                                */
/**************************************************************************/


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
IRQRETTYPE profibus_irq( int irq, void *dev_id, struct pt_regs *regs )
#else
IRQRETTYPE profibus_irq( int irq, void *dev_id )
#endif
{
  uchar IrqType;
  unsigned long flags;
  struct CntrlExt *Board = (struct CntrlExt *)dev_id;
//   static unsigned int counter = 0;
  
//   counter++;

//   if(!(counter%10000))printk("profibus_irq\n");
  
  if(!(*((u16 *)Board->c_to_h_irq_val)))
  {
//     if(!(counter%10000)) printk("irq is not for me!\n");
    return IRQ_NONE;
  }
  
//   if(!(counter%10000)) printk("profibus_irq ok for channel%lu\n",Board->Channel);
  
  Board->NrIrqs++;
  GetSpinlock(Board, &flags, __LINE__);
  IrqType = reset_cntrl_irq( Board );
  ReleaseSpinlock(Board, flags, __LINE__);
  
  // printk("profibus_irq: IrqType=0x%x\n", IrqType);
  
  
  
  
  //PR_DEBUG( "profibus irq (0x%x, %p): ", IrqType, Board );

  switch( IrqType ) 
  {
  case DP_SLAVE_IO_REQ_IRQ:
    // keep the ISR as short as possible, do all copying in an
    // interruptible thread immediate after leaving the ISR.
    // The thread will close the cycle by calling irq_to_cntrl(...)
    // Acknowledgement of the IRQ is done in the IST.
    //PR_DEBUG("DP_SLAVE_IO_REQ_IRQ\n");
   
    scheduleTask(Board);
    break;
  case DP_SLAVE_IO_ACK_IRQ:
   // printk("profibus_irq: DP_SLAVE_IO_ACK_IRQ\n");
    PR_DEBUG(D_SLAVE_L3,"DP_SLAVE_IO_ACK_IRQ\n");
    break; // not used
  case DP_DATA_STOP_REQ_IRQ:
   // printk("profibus_irq: DP_DATA_STOP_REQ_IRQ\n");
    PR_DEBUG(D_SLAVE_L1,"DP_DATA_STOP_REQ_IRQ\n");
   
    Board->active_tsklt = PBFIRSTTASK;
    Board->Dp.dpState = DP_OP_MODE_STOP;
    //cmi_exit_dp( Board );
    //irq_to_cntrl( Board, DP_SLAVE_IO_REQ_IRQ );
    // XXX - Move the following function irq_to_cntrl to a thread !!!!
    GetSpinlock(Board, &flags, __LINE__);
    irq_to_cntrl( Board, DP_DATA_STOP_ACK_IRQ); // acknowledge by IRQ
    ReleaseSpinlock(Board, flags, __LINE__);
    break;
  case DP_DATA_STOP_ACK_IRQ:
   // printk("profibus_irq: DP_DATA_STOP_REQ_IRQ\n");
    PR_DEBUG(D_DATA_L1,"DP_DATA_STOP_ACK_IRQ\n");
    break;
  case REQ_IRQ:
  //  printk("profibus_irq: REQ_IRQ\n");
    PR_DEBUG(D_SERVICE_L3,"REQ_IRQ\n");
    wake_up_interruptible( &Board->Service.QRead );
    free_wait(&Board->Service.wait_for_read);
    break;
  case ACK_IRQ:
   // printk("profibus_irq: ACK_IRQ\n");
    PR_DEBUG(D_SERVICE_L3,"ACK_IRQ\n");
    wake_up_interruptible( &Board->Service.QWrite );
    free_wait(&Board->Service.wait_for_write);
    break;
  case DPS_SLAVE_INP_REQ_IRQ:
  //  printk("profibus_irq: DPS_SLAVE_INP_REQ_IRQ\n");
    PR_DEBUG(D_SLAVE_L3,"DPS_SLAVE_INP_REQ_IRQ\n");
    GetSpinlock(Board, &flags, __LINE__);
    irq_to_cntrl( Board, DPS_SLAVE_INP_ACK_IRQ );
    ReleaseSpinlock(Board, flags, __LINE__);
    break;
  case DPS_SLAVE_OUTP_REQ_IRQ:
    
     scheduleTask(Board);
    break;
  // not used
  case DPS_SLAVE_INP_ACK_IRQ:
  case DPS_SLAVE_OUTP_ACK_IRQ:
   // printk("profibus_irq: DPS_SLAVE_OUTP_ACK_IRQ\n");
    wake_up_interruptible( &Board->SlaveIn.QInputAtController );
    break;
  case EXCEPTION_REQ_IRQ:
   // printk("profibus_irq: EXCEPTION_REQ_IRQ\n");
    Board->Service.Exception = 1;
  //  printk("EXCEPTION %d\n", IrqType );
    wake_up_interruptible( &Board->Service.QRead );
    free_wait(&Board->Service.wait_for_read);
    wake_up_interruptible( &Board->Service.QWrite );
    free_wait(&Board->Service.wait_for_write);
    wake_up_interruptible( &Board->SlaveIn.QInputAtController );
    break;
  case EXCEPTION_ACK_IRQ:
  case TRC_REQ_IRQ:
  case TRC_ACK_IRQ:
  default:
   // printk("TRC_IRQ %d\n", IrqType );
    break;
  }
  
//   printk("profibus_irq finished!\n");
  return IRQ_HANDLED;
}



unsigned short reset_cntrl_irq( struct CntrlExt *CtrlExt)
{
  //char *DprAddr = (char *)CtrlExt->DprAddr;
  short volatile *c_to_h_irq_val;
  short irqVal;

  c_to_h_irq_val = CtrlExt->c_to_h_irq_val; //DprAddr + C_TO_H_IRQ_VAL_OFFSET;  // get IRQ identifier
  irqVal = read_io_word( c_to_h_irq_val );
  write_io_word( c_to_h_irq_val, 0x00 );
  return( irqVal );
}

void await_irq_reset( struct CntrlExt *CtrlExt)
{
  short volatile *h_to_c_irq_val = CtrlExt->h_to_c_irq_val; 
  int timeout;

  // printk("reset_cntrl_irq\n");
  //PR_DEBUG("await_irq_reset( 0x%p )\n", h_to_c_irq_val );
  StartTimeout( &timeout, CMI_TIMEOUT);
  
  // wait until controller has reset IRQ identifier or timeout is detected
  while( (read_io_byte( h_to_c_irq_val ) != 0x00) ) 
  {
    if( SleepAWhile(&timeout) )
      break;
  }
  
  // printk("reset_cntrl_irq finished!\n");
  return;
}


/*******************/
/* Debug Functions */
/*******************/

void DpramOut( struct CntrlExt *Board, int offset )
{
  char *pDpram = (char*)Board->DprAddr;
  
  DumpMem(pDpram+offset, 0x200, "DPRAM");

}







