/******************************************************************************
*                                                                             *
*                                 SOFTING AG                                  *
*                          Richard-Reitzner-Allee 6                           *
*                                D-85540 Haar                                 *
*                        Phone: (++49)-(0)89-45656-0                          *
*                        Fax:   (++49)-(0)89-45656-399                        *
*                                                                             *
*                      Copyright (C) SOFTING AG 1995-2003                     *
*                              All Rights Reserved                            *
*                                                                             *
*******************************************************************************

FILE_NAME               DPSSched.c

PROJECT_NAME            PROFIBUS

PROJECT_NAME            PROFIBUS DP-Slave

MODULE                  DPSDemo

COMPONENT_LIBRARY       --------

AUTHOR                  SOFTING AG

VERSION                 1.20

DATE                    30.04.2000

STATUS

FUNCTIONAL_MODULE_DESCRIPTION

DPS Demo application
------------------------------

The DPSDemo application shows you how to access the functionality of the PROFIBUS PAPI.
It uses the functions of the PROFIBUS PAPI to communicate with the DP slave firmware.

Program structure
-----------------

The following hierarchichal structure with six threads is used:


main (DPSFrame.C): - check the commandline, initialize hardware settings
|                  - starts the scheduler
|                  - waits for the end of one of the created tasks to quit the program
|
|-- dpsframe_init (DPSFrame.C): - initialize all modules, exit- and abnormal
|   |                             termination handler
    |                           - creates the keyboard polling task
    |
    |-- dpssched_init (DPSSched.C): - initialize the task-, event- and timer control
    |   |
    |   |-- system_timer_task (DPSSched.C): - receive the ticks from DOS timer
    |   |                                     interrupt (55 ms)
    |                                       - looks for timer expirations and
    |                                         will inform the respective tasks via
    |                                         timer event
    |
    |-- dpsprofi_init (DPSProfi.C): - initialize the hardware and the PAPI
    |   |                           - initialize simulation of input data
    |   |
    |   |-- papi_task (DPSProfi.C): - controls the PAPI service interface
    |   |                           - send request and responses ( DPS_Init_Slave,
    |   |                             DPS_Get_Status )
    |   |                           - handle all events ( DPS_Set_Prm, etc. )
    |   |
    |   |-- input_task (DPSProfi.C): - simulate the input data of the Slave-Application
    |   |                            - increments a counter asynchron, after receiving an
    |   |                              event from the DOS timer interrupt
    |   |                            - read synchron the value of the BIOS-Timer
    |   |
    |   |-- output_task (DPSProfi.C): - after reveiving the cyclic timer event from the
    |   |                               scheduler system timer, read this task the output
    |                                   data and displays it
    |
    |-- keyboard_task (DPSFrame.C): - handles the keyboard input
                                    - display the possible commands after keys h or space
                                    - send an event to papi_task after key g
                                    - terminates after keys q, x or ESC


RELATED_DOCUMENTS
=============================================================================*/

#include "keywords.h"

INCLUDES

#if defined (WIN16) || defined (WIN32)
#include <windows.h>
#include <conio.h>
#include <dos.h>
#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <unistd.h>

#include "pb_type.h"
#include "pb_err.h"

#include "dpssched.h"

/****************************************************************************/

LOCAL_DEFINES

/****************************************************************************/

LOCAL_TYPEDEFS

/*--- SCHEDULER INFORMATION ------------------------------------------------*/

typedef struct _T_DPSSCHED_INFO
{
  T_LIST_ELEMENT              list_element;

  T_DPSSCHED_TASK_PRIORITY    priority;

  USIGN32                     current_priority;
  USIGN32                     priority_increment;

} T_DPSSCHED_INFO;


/*--- TIMER INFORMATION ----------------------------------------------------*/

typedef struct _T_DPSSCHED_TIMER_INFO
{
  T_LIST_ELEMENT    list_element;

  USIGN32           duration;
  USIGN32           current_duration;

  PB_BOOL           cyclic;

} T_DPSSCHED_TIMER_INFO;


/*--- TASK OBJECT ----------------------------------------------------------*/

typedef struct _T_DPSSCHED_TASK_OBJECT
{
  T_DPSSCHED_TASK_ID      task_id;
  VOID                    (*task_function_ptr) (T_DPSSCHED_EVENT_ID);

  T_DPSSCHED_INFO         schedule_info;
  T_DPSSCHED_TIMER_INFO   timer_info;

  USIGN16                 event_counter [DPSSCHED_MAX_NUMBER_EVENTS];

} T_DPSSCHED_TASK_OBJECT;


/****************************************************************************/

FUNCTION_DECLARATIONS

  LOCAL VOID dpssched_insert_started_timer  (T_DPSSCHED_TIMER_INFO FAR* new_timer_info_ptr);


/****************************************************************************/

EXPORT_DATA

/****************************************************************************/

IMPORT_DATA

/****************************************************************************/

LOCAL_DATA

  /*--- SCHEDULING AND EVENT HANDLING --------------------------------------*/

  static        T_DPSSCHED_TASK_OBJECT    task_object [DPSSCHED_MAX_NUMBER_TASKS];

  static        T_LIST                    schedule_list;
  static        PB_BOOL                   schedule_stopped = PB_FALSE;

#if defined (WIN32) || defined (_WIN32) || defined (WIN16) || defined (_WIN16)
  /*--- TIMER INTERRUPT PROVIDES SYSTEM TIMER ------------------------------*/
  static        UINT                      dpssched_timer_id;
#endif

  static        T_LIST                    timer_list;
  static        USIGN32                   current_time = 0uL;

  static const  USIGN32                   timer_resolution = 8000; //55uL;
  pthread_mutex_t                  hMutex;


/****************************************************************************/
/*** DOUBLE LINKED LIST *****************************************************/
/****************************************************************************/

FUNCTION GLOBAL VOID list_element_init
  (
  IN    T_LIST_ELEMENT   FAR *element_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - initializes the referenced list element                              */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( element_ptr != NULL );

  element_ptr->next_ptr = NULL;
  element_ptr->prev_ptr = NULL;
  element_ptr->data_ptr = NULL;

} /* list_element_init */

/****************************************************************************/

FUNCTION GLOBAL VOID list_init
  (
  IN    T_LIST    FAR *list_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - initializes the linking of the list base elements                    */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( list_ptr != NULL );

  list_ptr->head.next_ptr   = &list_ptr->tail;
  list_ptr->head.prev_ptr   = &list_ptr->head;
  list_ptr->head.data_ptr   = NULL;

  list_ptr->tail.next_ptr   = &list_ptr->tail;
  list_ptr->tail.prev_ptr   = &list_ptr->head;
  list_ptr->tail.data_ptr   = NULL;

  list_ptr->iterator_ptr    = &list_ptr->head;

  list_ptr->number_elements = 0;

} /* list_init */

/****************************************************************************/

FUNCTION GLOBAL T_LIST_ELEMENT FAR* list_get_first
  (
  IN    T_LIST    FAR *list_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - returns the first element of the given list and resets the iterator  */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( list_ptr->head.next_ptr != NULL );

  list_ptr->iterator_ptr = list_ptr->head.next_ptr;

  if (list_ptr->iterator_ptr != &list_ptr->tail)
  {
    return (list_ptr->iterator_ptr);
  }

  return (NULL);

} /* list_get_first */

/****************************************************************************/

FUNCTION GLOBAL T_LIST_ELEMENT FAR* list_get_last
  (
  IN    T_LIST    FAR *list_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - returns the last element of the given list and resets the iterator   */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( list_ptr->tail.prev_ptr != NULL );

  list_ptr->iterator_ptr = list_ptr->tail.prev_ptr;

  if (list_ptr->iterator_ptr != &list_ptr->head)
  {
    return (list_ptr->iterator_ptr);
  }

  return (NULL);

} /* list_get_last */

/****************************************************************************/

FUNCTION GLOBAL T_LIST_ELEMENT FAR* list_get_next
  (
  IN    T_LIST    FAR *list_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - returns the next available element of the double linked list         */
  /*   which is referenced by the iterator_ptr                              */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( list_ptr != NULL );

  list_ptr->iterator_ptr = list_ptr->iterator_ptr->next_ptr;

  if (list_ptr->iterator_ptr != &list_ptr->tail)
  {
    return (list_ptr->iterator_ptr);
  }

  return (NULL);

} /* list_get_next */

/****************************************************************************/

FUNCTION GLOBAL T_LIST_ELEMENT FAR* list_get_prev
  (
  IN    T_LIST    FAR *list_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - returns the prev available element of the double linked list         */
  /*   which is referenced by the iterator_ptr                              */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( list_ptr != NULL );

  list_ptr->iterator_ptr = list_ptr->iterator_ptr->prev_ptr;

  if (list_ptr->iterator_ptr != &list_ptr->head)
  {
    return (list_ptr->iterator_ptr);
  }

  return (NULL);

} /* list_get_prev */

/****************************************************************************/

FUNCTION GLOBAL VOID list_add_head
  (
  IN    T_LIST           FAR *list_ptr,
  IN    T_LIST_ELEMENT   FAR *element_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - inserts the referenced element at the head of a double linked list   */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( element_ptr != NULL );
  assert( list_ptr->head.next_ptr != NULL );
  assert( element_ptr->next_ptr == NULL );
  assert( element_ptr->prev_ptr == NULL );

  element_ptr->next_ptr           = list_ptr->head.next_ptr;
  element_ptr->next_ptr->prev_ptr = element_ptr;
  list_ptr->head.next_ptr         = element_ptr;
  element_ptr->prev_ptr           = &list_ptr->head;

  list_ptr->number_elements++;

} /* list_add_head */

/****************************************************************************/

FUNCTION GLOBAL VOID list_add_tail
  (
  IN    T_LIST           FAR *list_ptr,
  IN    T_LIST_ELEMENT   FAR *element_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - inserts the referenced element at the end of a double linked list    */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( element_ptr != NULL );
  assert( list_ptr->tail.prev_ptr != NULL );

  element_ptr->prev_ptr           = list_ptr->tail.prev_ptr;
  element_ptr->prev_ptr->next_ptr = element_ptr;
  list_ptr->tail.prev_ptr         = element_ptr;
  element_ptr->next_ptr           = &list_ptr->tail;

  list_ptr->number_elements++;

} /* list_add_tail */

/****************************************************************************/

FUNCTION GLOBAL VOID list_add_before
  (
  IN    T_LIST           FAR *list_ptr,
  IN    T_LIST_ELEMENT   FAR *position_ptr,
  IN    T_LIST_ELEMENT   FAR *element_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - inserts the referenced element before the given position of a double */
  /*   linked list                                                          */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( element_ptr != NULL );
  assert( position_ptr != NULL );
  assert( position_ptr->next_ptr != NULL );
  assert( position_ptr->prev_ptr != NULL );


  element_ptr->next_ptr            = position_ptr;
  element_ptr->prev_ptr            = position_ptr->prev_ptr;

  position_ptr->prev_ptr->next_ptr = element_ptr;
  position_ptr->prev_ptr           = element_ptr;

  list_ptr->number_elements++;

} /* list_add_before */

/****************************************************************************/

FUNCTION GLOBAL VOID list_add_after
  (
  IN    T_LIST           FAR *list_ptr,
  IN    T_LIST_ELEMENT   FAR *position_ptr,
  IN    T_LIST_ELEMENT   FAR *element_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - inserts the referenced element after the given position of a double  */
  /*   linked list                                                          */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( element_ptr != NULL );
  assert( position_ptr != NULL );
  assert( element_ptr->next_ptr == NULL );
  assert( element_ptr->prev_ptr == NULL );
  assert( position_ptr->next_ptr != NULL );
  assert( position_ptr->prev_ptr != NULL );

  element_ptr->next_ptr            = position_ptr->next_ptr;
  element_ptr->prev_ptr            = position_ptr;

  position_ptr->next_ptr->prev_ptr = element_ptr;
  position_ptr->next_ptr           = element_ptr;

  list_ptr->number_elements++;

} /* list_add_after */

/****************************************************************************/

FUNCTION GLOBAL T_LIST_ELEMENT FAR* list_remove_head
  (
  IN    T_LIST    FAR *list_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - removes the element at the head of a double linked list              */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_LIST_ELEMENT    FAR *this_ptr;

FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( list_ptr->head.next_ptr != NULL );

  if (list_ptr->number_elements == 0)
  {
    return (NULL);
  }

  this_ptr                     = list_ptr->head.next_ptr;
  this_ptr->next_ptr->prev_ptr = &list_ptr->head;
  list_ptr->head.next_ptr      = this_ptr->next_ptr;

  #ifdef DEBUG
    this_ptr->next_ptr = NULL;
    this_ptr->prev_ptr = NULL;
  #endif /* DEBUG */

  list_ptr->number_elements--;

  return (this_ptr);

} /* list_remove_head */

/****************************************************************************/

FUNCTION GLOBAL T_LIST_ELEMENT FAR* list_remove_tail
  (
  IN    T_LIST    FAR *list_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - removes the element at the end of a double linked list               */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_LIST_ELEMENT    FAR *this_ptr;

FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( list_ptr->tail.prev_ptr != NULL );

  if (list_ptr->number_elements == 0)
  {
    return (NULL);
  }

  this_ptr                     = list_ptr->tail.prev_ptr;
  this_ptr->prev_ptr->next_ptr = &list_ptr->tail;
  list_ptr->tail.prev_ptr      = this_ptr->prev_ptr;

  #ifdef DEBUG
    this_ptr->next_ptr = NULL;
    this_ptr->prev_ptr = NULL;
  #endif /* DEBUG */

  list_ptr->number_elements--;

  return (this_ptr);

} /* list_remove_tail */

/****************************************************************************/

FUNCTION GLOBAL VOID list_remove_at
  (
  IN    T_LIST            FAR *list_ptr,
  IN    T_LIST_ELEMENT    FAR *element_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - removes the referenced element from the given list                   */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( element_ptr != NULL );
  assert( list_find( list_ptr, element_ptr ) != PB_FALSE );
  assert( element_ptr->prev_ptr != NULL );
  assert( element_ptr->next_ptr != NULL );

  element_ptr->next_ptr->prev_ptr = element_ptr->prev_ptr;
  element_ptr->prev_ptr->next_ptr = element_ptr->next_ptr;

  #ifdef DEBUG
    element_ptr->next_ptr = NULL;
    element_ptr->prev_ptr = NULL;
  #endif /* DEBUG */

  list_ptr->number_elements--;

} /* list_remove_at */

/****************************************************************************/

FUNCTION GLOBAL PB_BOOL list_find
  (
  IN    T_LIST            FAR *list_ptr,
  IN    T_LIST_ELEMENT    FAR *element_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - checks whether the referenced list element is contained in the list  */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_LIST_ELEMENT    FAR* iterator_ptr;


FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( element_ptr != NULL );
  assert( list_ptr->head.next_ptr != NULL );

  iterator_ptr = list_ptr->head.next_ptr;

  while (iterator_ptr != &list_ptr->tail)
  {
    if (iterator_ptr == element_ptr)
    {
      return (PB_TRUE);
    }

    assert( iterator_ptr->next_ptr != NULL );

    iterator_ptr = iterator_ptr->next_ptr;

  } /* while elements */

  return (PB_FALSE);

} /* list_find */

/****************************************************************************/

FUNCTION GLOBAL PB_BOOL list_find_remove
  (
  IN    T_LIST            FAR *list_ptr,
  IN    T_LIST_ELEMENT    FAR *element_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - checks whether the referenced list element is contained in the list  */
  /* - removes the element when found                                       */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_LIST_ELEMENT    FAR* iterator_ptr;


FUNCTION_BODY

  assert( list_ptr != NULL );
  assert( element_ptr != NULL );
  assert( list_ptr->head.next_ptr != NULL );

  iterator_ptr = list_ptr->head.next_ptr;

  while (iterator_ptr != &list_ptr->tail)
  {
    if (iterator_ptr == element_ptr)
    {

      assert( element_ptr->prev_ptr != NULL );
      assert( element_ptr->next_ptr != NULL );

      element_ptr->next_ptr->prev_ptr = element_ptr->prev_ptr;
      element_ptr->prev_ptr->next_ptr = element_ptr->next_ptr;

      #ifdef DEBUG
        element_ptr->next_ptr = NULL;
        element_ptr->prev_ptr = NULL;
      #endif /* DEBUG */

      list_ptr->number_elements--;

      return (PB_TRUE);
    }

    assert( iterator_ptr->next_ptr != NULL );

    iterator_ptr = iterator_ptr->next_ptr;

  } /* while elements */

  return (PB_FALSE);

} /* list_find_remove */

/****************************************************************************/
/*** SCHEDULER FUNCTIONS ****************************************************/
/****************************************************************************/

FUNCTION GLOBAL VOID dpssched_create_task
  (
  IN    T_DPSSCHED_TASK_ID          task_id,
  IN    T_DPSSCHED_TASK_PRIORITY    task_priority,
  IN    VOID                        (*task_function_ptr) (T_DPSSCHED_EVENT_ID)
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function insert a new task into the tasklist.                   */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_DPSSCHED_TASK_OBJECT FAR*   task_object_ptr;


FUNCTION_BODY

  assert (task_id < DPSSCHED_MAX_NUMBER_TASKS);
  assert (task_priority <= DPSSCHED_MAX_NUMBER_PRIORITIES);
  assert (task_function_ptr != NULL);

  task_object_ptr = &task_object [task_id];
  assert (task_object_ptr->task_id == DPSSCHED_INVALID_TASK_ID);

  /*------------------------------------------------------------------------*/

  task_object_ptr->task_id           = task_id;
  task_object_ptr->task_function_ptr = task_function_ptr;

  task_object_ptr->schedule_info.priority = task_priority;

  if (task_priority != DPSSCHED_NO_SCHEDULE)
  {
    list_add_tail (&schedule_list, &task_object_ptr->schedule_info.list_element);
  }

  return;

} /* dpssched_create_task */

/****************************************************************************/

FUNCTION GLOBAL VOID dpssched_start_schedule
  (
  IN    VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function forces the scheduler to start the schedule loop.       */
  /* - First the function check the priority of all tasks in the tasklist.  */
  /*   Task with a high priority will be called frequently than tasks with  */
  /*   a low priority.                                                      */
  /* - Every schedule rotation begin with a check for asynchrone events for */
  /*   each event-type for each task. If an event is found the scheduler    */
  /*   process one event per type and task.                                 */
  /* - Following check the next current priority for polling. Compare for   */
  /*   each task the task-priority with the current priority. If the        */
  /*   priority of the task is equal, process this task.                    */
  /* - At the end of each schedule rotation change the current priority for */
  /*   next polling cycle.                                                  */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  const USIGN32                 common_factor = 10uL * 9uL * 8uL * 7uL * 6uL;
  USIGN32                       sum_priority;
  USIGN32                       lowest_priority;

  T_DPSSCHED_TASK_OBJECT FAR*   task_object_ptr;
  T_DPSSCHED_INFO FAR*          schedule_info_ptr;

  T_DPSSCHED_TASK_ID            task_id;
  T_DPSSCHED_EVENT_ID           event_id;
  USIGN16                       event_counter;


FUNCTION_BODY

    
  assert (DPSSCHED_MAX_NUMBER_PRIORITIES == 10);

  /*------------------------------------------------------------------------*/
  /*--- PREPARE SCHEDULING -------------------------------------------------*/
  /*------------------------------------------------------------------------*/

  schedule_info_ptr = (T_DPSSCHED_INFO FAR*) list_get_first (&schedule_list);
  sum_priority      = 0uL;

  while (schedule_info_ptr != NULL)
  {
    assert ( (schedule_info_ptr->priority != DPSSCHED_NO_SCHEDULE          ) &&
             (schedule_info_ptr->priority <= DPSSCHED_MAX_NUMBER_PRIORITIES) );

    sum_priority += (USIGN32) (schedule_info_ptr->priority);

    schedule_info_ptr = (T_DPSSCHED_INFO FAR*) list_get_next (&schedule_list);
  }

  sum_priority *= common_factor;

  /*------------------------------------------------------------------------*/

  schedule_info_ptr = (T_DPSSCHED_INFO FAR*) list_get_first (&schedule_list);

  while (schedule_info_ptr != NULL)
  {
    schedule_info_ptr->priority_increment = sum_priority / schedule_info_ptr->priority;
    schedule_info_ptr->current_priority   = schedule_info_ptr->priority_increment;

    schedule_info_ptr = (T_DPSSCHED_INFO FAR*) list_get_next (&schedule_list);
  }

  /*------------------------------------------------------------------------*/
  /*--- HANDLE SCHEDULING --------------------------------------------------*/
  /*------------------------------------------------------------------------*/

  schedule_stopped = PB_FALSE;

  while (! schedule_stopped)
  {
    /*----------------------------------------------------------------------*/
    /*--- EVENT HANDLING ---------------------------------------------------*/
    /*----------------------------------------------------------------------*/
    pthread_mutex_lock(&hMutex);
    for (task_id = 0; task_id < DPSSCHED_MAX_NUMBER_TASKS; task_id++)
    {
      task_object_ptr = &task_object [task_id];

      
      for (event_id = DPSSCHED_SCHEDULE_EVENT + 1; event_id < DPSSCHED_MAX_NUMBER_EVENTS; event_id++)
      {
        
        event_counter = task_object_ptr->event_counter [event_id];
        
        
        if (event_counter > 0)
        {
          
          task_object_ptr->event_counter [event_id]--;
          
        }
        
        
        
        if (event_counter > 0)
        {
          assert (task_object_ptr->task_function_ptr != NULL);
         // printf("dpssched_start_schedule: calling event %u task %u ..", event_id, task_id);
          (*task_object_ptr->task_function_ptr) (event_id);
          //printf(".ok!\n");

        } /* if event found */
      } /* for event_id */
      
      
    } /* for task_id */
    

    /*----------------------------------------------------------------------*/
    /*--- SCHEDULE HANDLING ------------------------------------------------*/
    /*----------------------------------------------------------------------*/

    /*--- SEARCH LOWEST CURRENT PRIORITY FOR THIS SCHEDULE ROTATION --------*/

    schedule_info_ptr = (T_DPSSCHED_INFO FAR*) list_get_first (&schedule_list);
    lowest_priority   = (USIGN32) -1l;     /* start value: largest priority */

    while (schedule_info_ptr != NULL)
    {
      if (schedule_info_ptr->current_priority < lowest_priority)
      {
        lowest_priority = schedule_info_ptr->current_priority;
      }

      schedule_info_ptr = (T_DPSSCHED_INFO FAR*) list_get_next (&schedule_list);
    }

    if (lowest_priority != (USIGN32) -1l)      /* scheduling task available */
    {
      /*--- EXECUTE TASK WITH LOWEST CURRENT PRIORITY ----------------------*/

      schedule_info_ptr = (T_DPSSCHED_INFO FAR*) list_get_first (&schedule_list);

      while (schedule_info_ptr != NULL)
      {
        if (schedule_info_ptr->current_priority == lowest_priority)
        {
          task_object_ptr = (T_DPSSCHED_TASK_OBJECT FAR*) schedule_info_ptr->list_element.data_ptr;
          assert (task_object_ptr != NULL);
          assert (task_object_ptr->task_function_ptr != NULL);

          (*task_object_ptr->task_function_ptr) (DPSSCHED_SCHEDULE_EVENT);

          schedule_info_ptr->current_priority += schedule_info_ptr->priority_increment;
        }

        schedule_info_ptr = (T_DPSSCHED_INFO FAR*) list_get_next (&schedule_list);
      }

      /*--- PREPARE NEW SCHEDULE CYCLE -------------------------------------*/

      assert (lowest_priority <= sum_priority);

      if (lowest_priority == sum_priority)
      {
        schedule_info_ptr = (T_DPSSCHED_INFO FAR*) list_get_first (&schedule_list);

        while (schedule_info_ptr != NULL)
        {
          schedule_info_ptr->current_priority = schedule_info_ptr->priority_increment;

          schedule_info_ptr = (T_DPSSCHED_INFO FAR*) list_get_next (&schedule_list);
        }
      } /* if last schedule cycle */
    } /* if scheduling tasks */
    pthread_mutex_unlock(&hMutex);
    // please schedule!
    usleep(1);

  } /* while not schedule_stopped */

  
  return;

} /* dpssched_start_schedule */

/****************************************************************************/

FUNCTION GLOBAL VOID dpssched_stop_schedule
  (
  IN    VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function forces the scheduler to leave the schedule loop.       */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  schedule_stopped = PB_TRUE;

  return;

} /* dpssched_stop_schedule */

/****************************************************************************/

#if defined (WIN32) || defined (_WIN32) || defined (WIN16) || defined (_WIN16)
#pragma check_stack (off)
#endif

FUNCTION GLOBAL VOID dpssched_send_event
  (
  IN    T_DPSSCHED_TASK_ID    task_id,
  IN    T_DPSSCHED_EVENT_ID   event_id
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function indicates the given event to the specified task.       */
  /*   The task will be scheduled in the following to receive and handle    */
  /*   the event with highest priority.                                     */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_DPSSCHED_TASK_OBJECT FAR*   task_object_ptr;


FUNCTION_BODY

    
  assert (task_id < DPSSCHED_MAX_NUMBER_TASKS);
  assert (event_id < DPSSCHED_MAX_NUMBER_EVENTS);

  task_object_ptr = &task_object [task_id];
  assert (task_object_ptr->task_id != DPSSCHED_INVALID_TASK_ID);

  /*------------------------------------------------------------------------*/
  
  task_object_ptr->event_counter [event_id]++;
    assert (task_object_ptr->event_counter [event_id] > 0);
  
  return;

} /* dpssched_send_event */

#if defined (WIN32) || defined (_WIN32) || defined (WIN16) || defined (_WIN16)
#pragma check_stack
#endif

/****************************************************************************/

FUNCTION GLOBAL VOID dpssched_start_timer
  (
  IN    T_DPSSCHED_TASK_ID    task_id,
  IN    USIGN32               duration,
  IN    PB_BOOL               cyclic
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function starts the task timer of the given task_id.            */
  /* - The duration of the timer has an internal resolution of 1ms and is   */
  /*   mapped to the DOS system timer interrupt of 55ms.                    */
  /* - If the timer is already running then the timer will be restarted.    */
  /* - The parameter 'cyclic' detemines whether the timer will be a one     */
  /*   shot timer or a cyclic timer.                                        */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_DPSSCHED_TIMER_INFO FAR*    new_timer_info_ptr;


FUNCTION_BODY

  assert (task_id < DPSSCHED_MAX_NUMBER_TASKS);

  /*--- RESTART RUNNING TIMER ----------------------------------------------*/

  dpssched_stop_timer (task_id);

  /*--- PLACE TIMER IN TIMER QUEUE (EXPIRATION TIME SORTED) ----------------*/

  new_timer_info_ptr = &task_object [task_id].timer_info;

  assert (new_timer_info_ptr->duration == 0uL);
  assert (new_timer_info_ptr->current_duration == 0uL);
  assert (new_timer_info_ptr->cyclic == PB_FALSE);

  if (duration <= timer_resolution)
  {
    duration = timer_resolution + 1uL;    /* wait at minimum one timer tick */
  }

  new_timer_info_ptr->duration         = duration;
  new_timer_info_ptr->current_duration = duration;
  new_timer_info_ptr->cyclic           = cyclic;

  dpssched_insert_started_timer (new_timer_info_ptr);

  return;

} /* dpssched_start_timer */

/****************************************************************************/

FUNCTION GLOBAL VOID dpssched_stop_timer
  (
  IN    T_DPSSCHED_TASK_ID    task_id
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function stops the task timer of the given task.                */
  /* - If the timer is already expired any pending events are cleared.      */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_DPSSCHED_TASK_OBJECT FAR*   task_object_ptr;


FUNCTION_BODY

  assert ( (task_id  < DPSSCHED_MAX_NUMBER_TASKS) &&
           (task_id != DPSSCHED_TIMER_TASK_ID   ) );

  task_object_ptr = &task_object [task_id];

  /*--- CHECK IF TIMER ALREADY STARTED OR EXPIRED --------------------------*/

  if (! list_find_remove (&timer_list, &task_object_ptr->timer_info.list_element))
  {
    task_object_ptr->event_counter [DPSSCHED_TIMER_EVENT] = 0;
  }

#ifdef DEBUG
  else
  {
    task_object_ptr->timer_info.duration         = 0uL;
    task_object_ptr->timer_info.current_duration = 0uL;
    task_object_ptr->timer_info.cyclic           = PB_FALSE;

  } /* else timer not started */
#endif

  return;

} /* dpssched_stop_timer */

/****************************************************************************/

FUNCTION LOCAL VOID dpssched_insert_started_timer
  (
  IN    T_DPSSCHED_TIMER_INFO FAR*    new_timer_info_ptr
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function inserts the already initialized timer into the         */
  /*   timer_list.                                                          */
  /* - The timer_list is a sorted list which contains the soonest           */
  /*   expirations at the beginning of the list.                            */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_DPSSCHED_TIMER_INFO FAR*    timer_info_ptr;


FUNCTION_BODY

  assert (new_timer_info_ptr != NULL);

  /*--- NORMALIZE RUNNING TIMER (NEW START POINT) --------------------------*/

  timer_info_ptr = (T_DPSSCHED_TIMER_INFO FAR*) list_get_first (&timer_list);

  while (timer_info_ptr != NULL)
  {
    assert (timer_info_ptr->current_duration >= current_time);

    timer_info_ptr->current_duration -= current_time;
    timer_info_ptr = (T_DPSSCHED_TIMER_INFO FAR*) list_get_next (&timer_list);
  }

  current_time = 0uL;

  /*--- PLACE TIMER IN TIMER QUEUE (EXPIRATION TIME SORTED) ----------------*/

  assert (new_timer_info_ptr->duration > timer_resolution);
  assert (new_timer_info_ptr->current_duration == new_timer_info_ptr->duration);

  timer_info_ptr = (T_DPSSCHED_TIMER_INFO FAR*) list_get_first (&timer_list);

  while (timer_info_ptr != NULL)
  {
    if (timer_info_ptr->current_duration > new_timer_info_ptr->current_duration)
    {
      break;
    }

    timer_info_ptr = (T_DPSSCHED_TIMER_INFO FAR*) list_get_next (&timer_list);

  } /* while timer_list */

  if (timer_info_ptr != NULL)
  {
    list_add_before (&timer_list, &timer_info_ptr->list_element, &new_timer_info_ptr->list_element);
  }
  else
  {
    list_add_tail (&timer_list, &new_timer_info_ptr->list_element);
  }

  return;

} /* dpssched_insert_started_timer */

/****************************************************************************/

FUNCTION LOCAL VOID system_timer_task
  (
  IN    T_DPSSCHED_EVENT_ID    event_id
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function receives the timer ticks from the DOS system timer     */
  /*   interrupt (55ms).                                                    */
  /* - The function looks for timer expirations and will inform the         */
  /*   respective tasks via timer event.                                    */
  /* - Cyclic timers are restarted when they expire immediately.            */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  T_DPSSCHED_TIMER_INFO FAR*    timer_info_ptr;
  T_DPSSCHED_TASK_OBJECT FAR*   task_object_ptr;

  T_LIST                        restart_timer_list;


FUNCTION_BODY

  assert (event_id == DPSSCHED_TIMER_EVENT);

  list_init (&restart_timer_list);

  /*--- ADD NEW TIMER TICK -------------------------------------------------*/

  current_time += timer_resolution;

  /*--- CHECK TIMER LIST FOR EXPIRATIONS -----------------------------------*/

  timer_info_ptr = (T_DPSSCHED_TIMER_INFO FAR*) list_get_first (&timer_list);

  while (timer_info_ptr != NULL)
  {
    if (timer_info_ptr->current_duration > current_time)
    {
      break;      /* no timer expiration from here to the end of timer_list */
    }

    /*--- TIMER HAS EXPIRED ------------------------------------------------*/

    (VOID) list_remove_head (&timer_list);

    task_object_ptr = (T_DPSSCHED_TASK_OBJECT FAR*) timer_info_ptr->list_element.data_ptr;
    assert (task_object_ptr != NULL);

    dpssched_send_event (task_object_ptr->task_id, DPSSCHED_TIMER_EVENT);

    if (timer_info_ptr->cyclic)
    {
      timer_info_ptr->current_duration = timer_info_ptr->duration;

      list_add_tail (&restart_timer_list, &timer_info_ptr->list_element);

    } /* if cyclic timer */

#ifdef DEBUG
    else
    {
      timer_info_ptr->duration         = 0uL;
      timer_info_ptr->current_duration = 0uL;
      timer_info_ptr->cyclic           = PB_FALSE;

    } /* else cyclic timer */
#endif

    /*--- CHECK NEXT TIMER FOR EXPIRATION (THIS IS NOW THE FIRST!) ---------*/

    timer_info_ptr = (T_DPSSCHED_TIMER_INFO FAR*) list_get_first (&timer_list);

  } /* while timer_list */

  /*--- RESTART PENDING CYCLIC TIMERS --------------------------------------*/

  timer_info_ptr = (T_DPSSCHED_TIMER_INFO FAR*) list_remove_head (&restart_timer_list);

  while (timer_info_ptr != NULL)
  {
    dpssched_insert_started_timer (timer_info_ptr);

    timer_info_ptr = (T_DPSSCHED_TIMER_INFO FAR*) list_remove_head (&restart_timer_list);
  }

  return;

} /* system_timer_task */

/****************************************************************************/

#if defined (WIN32) || defined (_WIN32) || defined (WIN16) || defined (_WIN16)
#pragma check_stack (off)

FUNCTION VOID CALLBACK dpssched_system_timer_proc
  (
  IN UINT   uID,
  IN UINT   uMsg,
  IN DWORD  dwUser,
  IN DWORD  dw1,
  IN DWORD  dw2
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - This function is used to retrieve one software timer for each        */
  /*   task of the scheduler.                                               */
  /* - The timer interrupt service routine is chained with the DOS system   */
  /*   timer interrupt (55ms) and mapped to a internal base of 1ms.         */
  /* - The ISR retriggers the timer task which is necessary to provide      */
  /*   mutual exlusion during access of the timer queue.                    */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

  /*--- STIMULATE TIMER TASK - DO NOT ACCESS TIMER QUEUE WITIHIN IRQ -------*/

  dpssched_send_event (DPSSCHED_TIMER_TASK_ID, DPSSCHED_TIMER_EVENT);

  return;

} /* dpssched_system_timer_proc */

#pragma check_stack
#endif

/****************************************************************************/
/*** INIT / EXIT MODULE *****************************************************/
/****************************************************************************/

FUNCTION PUBLIC PB_BOOL dpssched_init
  (
  IN    VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - initializes this module                                              */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

  int      task_index;

#if defined (WIN32) || defined (_WIN32) || defined (WIN16) || defined (_WIN16)
  MMRESULT mmresult;
#endif

FUNCTION_BODY

  /*--- TASK OBJECTS -------------------------------------------------------*/

  memset (task_object, 0, sizeof (task_object));

  list_init (&schedule_list);

  for (task_index = 0; task_index < DPSSCHED_MAX_NUMBER_TASKS; task_index++)
  {
    list_element_init (&task_object [task_index].schedule_info.list_element);
    list_element_init (&task_object [task_index].timer_info.list_element);

    task_object [task_index].task_id = DPSSCHED_INVALID_TASK_ID;

    task_object [task_index].schedule_info.list_element.data_ptr = &task_object [task_index];
    task_object [task_index].timer_info.list_element.data_ptr    = &task_object [task_index];
  }

  /*--- INSTALL SYSTEM TIMER -----------------------------------------------*/

  list_init (&timer_list);

  dpssched_create_task (DPSSCHED_TIMER_TASK_ID, DPSSCHED_NO_SCHEDULE, system_timer_task);

#if defined (WIN32) || defined (_WIN32) || defined (WIN16) || defined (_WIN16)
  /* --- START 55ms TIMER ------------------------------------------------- */
  mmresult = timeSetEvent(55,55,(LPTIMECALLBACK) dpssched_system_timer_proc,0,TIME_PERIODIC);
  dpssched_timer_id = (UINT) mmresult;

  if   (mmresult) return(PB_TRUE);
  else            return(PB_FALSE);
#endif

#ifdef _LINUX
  return(PB_TRUE);
#endif

} /* dpssched_init */

/****************************************************************************/

FUNCTION PUBLIC VOID dpssched_exit
  (
  IN    VOID
  )
  /*------------------------------------------------------------------------*/
  /* FUNCTIONAL_DESCRIPTION                                                 */
  /*------------------------------------------------------------------------*/
  /* - terminates this module                                               */
  /*------------------------------------------------------------------------*/
{
LOCAL_VARIABLES

FUNCTION_BODY

#if defined (WIN32) || defined (_WIN32) || defined (WIN16) || defined (_WIN16)
  /*--- KILL DPSSPROFI TIMER ------------------------------------------------*/
  timeKillEvent(dpssched_timer_id);
#endif

  return;

} /* dpssched_exit */

/*** END OF DPSSched.c ******************************************************/
